# -*- coding: utf-8 -*-
"""Implements core image handling classes for the :mod:`Stoner.Image` package."""
__all__ = ["ImageArray", "ImageFile", "DrawProxy", "MaskProxy"]
import numpy as np
import os
import warnings
from copy import copy, deepcopy
from skimage import (
    color,
    exposure,
    feature,
    io,
    measure,
    filters,
    graph,
    util,
    restoration,
    morphology,
    segmentation,
    transform,
    viewer,
    draw,
)
from PIL import Image
from PIL import PngImagePlugin  # for saving metadata
import matplotlib.pyplot as plt
from Stoner.Core import typeHintedDict, metadataObject, regexpDict, DataFile
from Stoner.Image.util import convert
from Stoner import Data
from Stoner.tools import istuple, fix_signature, islike_list, get_option
from Stoner.compat import (
    python_v3,
    string_types,
    get_filedialog,
    bytes,
    int_types,
)  # Some things to help with Python2 and Python3 compatibility
import inspect
from functools import wraps

if python_v3:
    from io import BytesIO as StreamIO
else:
    from cStringIO import StringIO as StreamIO


dtype_range = {
    np.bool_: (False, True),
    np.bool8: (False, True),
    np.uint8: (0, 255),
    np.uint16: (0, 65535),
    np.int8: (-128, 127),
    np.int16: (-32768, 32767),
    np.int64: (-2 ** 63, 2 ** 63 - 1),
    np.uint64: (0, 2 ** 64 - 1),
    np.int32: (-2 ** 31, 2 ** 31 - 1),
    np.uint32: (0, 2 ** 32 - 1),
    np.float16: (-1, 1),
    np.float32: (-1, 1),
    np.float64: (-1, 1),
}


class ImageArray(np.ma.MaskedArray, metadataObject):

    """:py:class:`Stoner.Image.core.ImageArray` is a numpy array like class with a metadata parameter and pass through to skimage methods.

    ImageArray is for manipulating images stored as a 2d numpy array.
    It is built to be almost identical to a numpy array except for one extra
    parameter which is the metadata. This stores information about the image
    in a dictionary object for later retrieval.
    All standard numpy functions should work as normal and casting two types
    together should yield a ImageArray type (ie. ImageArray+np.ndarray=ImageArray)

    In addition any function from skimage should work and return a ImageArray.
    They can be called as eg. im=im.gaussian(sigma=2). Don't include the module
    name, just the function name (ie not filters.gaussian). Also omit the first
    image argument required by skimage.

    Attributes:
        metadata (dict):
            dictionary of metadata for the image
        clone (self):
            copy of self
        max_box (tuple):
            coordinate extent (xmin,xmax,ymin,ymax)


    For clarity it should be noted that any function will not alter the current
    instance, it will clone it first then return the clone after performing the
    function on it.

   Note:

        For arrays the indexing is (row, column). However the normal way to index
        an image would be to do (horizontal, vert), which is the opposite.
        In ImageArray the coordinate system is chosen similar to skimage. y points
        down x points right and the origin is in the top left corner of the image.
        When indexing the array therefore you need to give it (y,x) coordinates
        for (row, column).::

             ----> x (column)
            |
            |
            v
            y (row)

        eg I want the 4th pixel in the horizontal direction and the 10th pixel down
        from the top I would ask for ImageArray[10,4]

        but if I want to translate the image 4 in the x direction and 10 in the y
        I would call im=im.translate((4,10))

    """

    # Proxy attributess for storing imported functions. Only do the import when needed
    _func_proxy = None

    # extra attributes for class beyond standard numpy ones
    _extra_attributes_default = {"metadata": typeHintedDict({}), "filename": ""}

    # Default values for when we can't find the attribute already
    _defaults = {"debug": False, "_hardmask": False}

    fmts = ["png", "npy", "tiff", "tif"]

    # now initialise class

    if not python_v3:  # Ugh what a horrible hack!
        _mask = np.ma.MaskedArray([]).mask

    def __new__(cls, *args, **kargs):
        """Construct an ImageArray object.

        We're using __new__ rather than __init__ to imitate a numpy array as
        close as possible.
        """

        if len(args) not in [0, 1]:
            raise ValueError("ImageArray expects 0 or 1 arguments, {} given".format(len(args)))

        ### Deal with kwargs
        array_arg_keys = ["dtype", "copy", "order", "subok", "ndmin", "mask"]  # kwargs for array setup
        array_args = {k: kargs.pop(k) for k in array_arg_keys if k in kargs.keys()}
        user_metadata = kargs.pop("metadata", {})
        asfloat = kargs.pop("asfloat", False) or kargs.pop(
            "convert_float", False
        )  # convert_float for back compatability
        _debug = kargs.pop("debug", False)

        ### 0 args initialisation
        if len(args) == 0:
            ret = np.empty((0, 0), dtype=float).view(cls)
        else:

            ### 1 args initialisation
            arg = args[0]
            loadfromfile = False
            if isinstance(arg, cls):
                ret = arg
            elif isinstance(arg, np.ndarray):
                # numpy array or ImageArray)
                if len(arg.shape) == 2:
                    ret = arg.view(ImageArray)
                elif len(arg.shape) == 1:
                    arg = np.array([arg])  # force 2d array
                    ret = arg.view(ImageArray)
                else:
                    raise ValueError("Array dimension 0 must be at most 2. Got {}".format(len(arg.shape)))
                ret.metadata = getattr(arg, "metadata", typeHintedDict())
            elif isinstance(arg, bool) and not arg:
                patterns = (("png", "*.png"), ("npy", "*.npy"))
                arg = get_filedialog(what="r", filetypes=patterns)
                if len(arg) == 0:
                    raise ValueError("No file given")
                else:
                    loadfromfile = True
            elif isinstance(arg, string_types) or loadfromfile:
                # Filename- load datafile
                if not os.path.exists(arg):
                    raise ValueError("File path does not exist {}".format(arg))
                ret = cls._load(arg, **array_args)
            elif isinstance(arg, ImageFile):
                # extract the image
                ret = arg.image
            else:
                try:  # try converting to a numpy array (eg a list type)
                    ret = np.asarray(arg, **array_args).view(cls)
                    if ret.dtype == "O":  # object dtype - can't deal with this
                        raise ValueError
                except ValueError:  # ok couldn't load from iterable, we're done
                    raise ValueError("No constructor for {}".format(type(arg)))
            if asfloat and ret.dtype.kind != "f":  # convert to float type in place
                meta = ret.metadata  # preserve any metadata we may already have
                ret = convert(ret, np.float64).view(ImageArray)
                ret.metadata.update(meta)

        # all constructors call array_finalise so metadata is now initialised
        if "Loaded from" not in ret.metadata.keys():
            ret.metadata["Loaded from"] = ""
        ret.filename = ret.metadata["Loaded from"]
        ret.metadata.update(user_metadata)
        ret.debug = _debug
        return ret

    def __array_finalize__(self, obj):
        """__array_finalize__ and __array_wrap__ are necessary functions when subclassing numpy.ndarray to fix some behaviours.

        See http://docs.scipy.org/doc/numpy-1.10.1/user/basics.subclassing.html for
        more info and examples
        Defaults below are only set when constructing an array using view
        eg np.arange(10).view(ImageArray). Otherwise filename and metadata
        attributes are just copied over (plus any other attributes set in
        _optinfo).
        """
        if getattr(self, "debug", False):
            curframe = inspect.currentframe()
            calframe = inspect.getouterframes(curframe, 2)
        _extra_attributes = getattr(obj, "_optinfo", deepcopy(ImageArray._extra_attributes_default))
        setattr(self, "_optinfo", copy(_extra_attributes))
        for k, v in _extra_attributes.items():
            try:
                setattr(self, k, getattr(obj, k, v))
            except AttributeError:  # Some versions of  python don't like this
                pass
        super(ImageArray, self).__array_finalize__(obj)

    def __array_prepare__(self, arr, context=None):
        return super(ImageArray, self).__array_prepare__(arr, context)

    def __array_wrap__(self, out_arr, context=None):
        """Part of the numpy array machinery.

        see __array_finalize__ for info. This is for if ImageArray is called
        via ufuncs. array_finalize is called after.
        """
        ret = super(ImageArray, self).__array_wrap__(out_arr, context)
        return ret

    def __init__(self, *args, **kwargs):
        """Constructor method for :py:class:`ImageArray`.

        various forms are recognised

        .. py:function:: ImageArray('filename')
            :noindex:

            Creates the new ImageArray object and then loads data from the
            given *filename*.

        .. py:function:: ImageArray(array)
            :noindex:

            Creates a new ImageArray object and assigns the *array*.

        .. py:function:: ImageArray(ImageFile)
            :noindex:

            Creates a new ImageArray object and assigns the *array* using
            the ImageFile.image property.


        .. py:function:: ImageArray(ImageArray)
            :noindex:

            Creates the new ImageArray object and initialises all data from the
            existing :py:class:`ImageArray` instance. This on the face of it does the same as
            the assignment operator, but is more useful when one or other of the
            ImageArray objects is an instance of a sub - class of ImageArray

        .. py:function:: ImageArray(bool)
            :noindex:

           if arg[0] is bool and False then open a file dialog to locate the
           file to open.

        Args:
            arg (positional arguments): an argument that matches one of the
                definitions above
        Keyword Arguments: All keyword arguments that match public attributes are
                used to set those public attributes eg metadata.

                asfloat(bool):
                    if True  and loading the image from file, convert the image
                    to float values between 0 and 1 (necessary for some forms
                    of processing)
        """
        super(ImageArray, self).__init__(*args, **kwargs)

    @classmethod
    def _load(cls, filename, **kargs):
        """Load an image from a file and return as a ImageArray."""
        fmt = kargs.pop("fmt", os.path.splitext(filename)[1][1:])
        handlers = {"npy": cls._load_npy, "png": cls._load_png, "tiff": cls._load_tiff, "tif": cls._load_tiff}
        if fmt not in handlers:
            raise ValueError("{} is not a recognised format for loading.".format(fmt))
        ret = handlers[fmt](filename, **kargs)
        return ret

    @classmethod
    def _load_npy(cls, filename, **kargs):
        """Load image data from a numpy file."""
        image = np.load(filename)
        image = np.array(image, **kargs).view(cls)
        image.metadata["Loaded from"] = os.path.realpath(filename)
        image.filename = os.path.realpath(filename)
        return image

    @classmethod
    def _load_png(cls, filename, **kargs):
        with Image.open(filename, "r") as img:
            image = np.asarray(img).view(cls)
            # Since skimage.img_as_float() looks at the dtype of the array when mapping ranges, it's important to make
            # sure that we're not using too many bits to store the image in. This is a bit of a hack to reduce the bit-depth...
            if np.issubdtype(image.dtype, np.integer):
                bits = np.ceil(np.log2(image.max()))
                if bits <= 8:
                    image = image.astype("uint8")
                elif bits <= 16:
                    image = image.astype("uint16")
                elif bits <= 32:
                    image = image.astype("uint32")
            for k in img.info:
                v = img.info[k]
                if "b'" in v:
                    v = v.strip(" b'")
                image.metadata[k] = v
        image.metadata["Loaded from"] = os.path.realpath(filename)
        image.filename = os.path.realpath(filename)
        return image

    @classmethod
    def _load_tiff(cls, filename, **kargs):
        metadict = typeHintedDict({})
        with Image.open(filename, "r") as img:
            image = np.asarray(img)
            tags = img.tag_v2
            if 270 in tags:
                from json import loads

                try:
                    metadata_string = tags[270]
                    metadata = loads(metadata_string)
                except Exception:
                    metadata = []
            else:
                metadata = []
            metadict.import_all(metadata)

            # OK now try and sort out the datatype before loading
        dtype = metadict.get(
            "ImageArray.dtype", None
        )  # if tif was previously saved by Stoner then dtype should have been added to the metadata
        # If we convert to float, it's important to make
        # sure that we're not using too many bits to store the image in.
        # This is a bit of a hack to reduce the bit-depth...
        if dtype is None:
            if np.issubdtype(image.dtype, np.integer):
                bits = np.ceil(np.log2(image.max()))
                if bits <= 8:
                    dtype = "uint8"
                elif bits <= 16:
                    dtype = "uint16"
                elif bits <= 32:
                    dtype = "uint32"
            else:
                dtype = np.dtype(image.dtype).name  # retain the loaded datatype
        try:
            image = image.astype(dtype)
        except TypeError:  # Python 2.7 can throw up a bad type error here
            pass
        image = image.view(cls)
        image.update(metadict)
        image.metadata["Loaded from"] = os.path.realpath(filename)
        image.filename = os.path.realpath(filename)
        return image

    def _box(self, *args, **kargs):
        """Construct and indexing tuple for selecting areas for cropping and boxing.

        The box can be specified as:

            - (int): a fixed number of pxiels is removed from all sides
            - (float): the central region of the image is selected
            - None: the whole image is selected
            - False: The user can select a region of interest
            - (iterable of length 4) - assumed to give 4 integers to describe a specific box
        """
        if len(args) == 0 and "box" in kargs.keys():
            args = kargs["box"]  # back compatability
        elif len(args) not in (1, 4):
            raise ValueError("box accepts 1 or 4 arguments, {} given.".format(len(args)))
        if len(args) == 1:
            box = args[0]
            if isinstance(box, bool) and not box:  # experimental
                print("Select crop area")
                box = self.draw_rectangle(box)
            elif islike_list(box) and len(box) == 4:  # Full box as a list
                box = [x for x in box]
            elif box is None:  # Whole image
                box = [0, self.shape[1], 0, self.shape[0]]
            elif isinstance(box, int):  # Take a border of n pixels out
                box = [box, self.shape[1] - box, box, self.shape[0] - box]
            elif isinstance(box, float):  # Keep the central fraction of the image
                box = [
                    round(self.shape[1] * box / 2),
                    round(self.shape[1] * (1 - box / 2)),
                    round(self.shape[1] * box / 2),
                    round(self.shape[1] * (1 - box / 2)),
                ]
                box = list([int(x) for x in box])
            else:
                raise ValueError("crop accepts tuple of length 4, {} given.".format(len(box)))
        else:
            box = list(args)
        for i, item in enumerate(box):  # replace None with max extent
            if isinstance(item, float):
                if i < 2:
                    box[i] = round(self.shape[1] * item)
                else:
                    box[i] = round(self.shape[0] * item)
            elif isinstance(item, int_types):
                pass
            elif item is None:
                box[i] = self.max_box[i]
            else:
                raise TypeError("Arguments for box should be floats, integers or None, not {}".format(type(item)))
        return slice(box[2], box[3]), slice(box[0], box[1])

    #################################################################################################
    ############ Properties #########################################################################

    @property
    def aspect(self):
        """Return the aspect ratio (width/height) of the image."""
        return float(self.shape[1]) / self.shape[0]

    @property
    def centre(self):
        return tuple(np.array(self.shape) / 2.0)

    @property
    def clone(self):
        """return a copy of the instance"""
        ret = ImageArray(np.copy(self))
        self._optinfo["mask"] = self.mask  # Make sure we've updated our mask record
        for k, v in self._optinfo.items():
            try:
                setattr(ret, k, deepcopy(v))
            except Exception:
                setattr(ret, k, copy(v))
        return ret

    @property
    def flat(self):
        """MaskedArray.flat doesn't work the same as array.flat."""
        return np.asarray(self).flat

    @property
    def max_box(self):
        """return the coordinate extent (xmin,xmax,ymin,ymax)"""
        box = (0, self.shape[1], 0, self.shape[0])
        return box

    #    @property
    #    def data(self):
    #        """alias for image[:]. Equivalence to Stoner.data behaviour"""
    #        return self[:]

    @property
    def CW(self):
        """Rotate clockwise by 90 deg."""
        return self.T[:, ::-1]

    @property
    def CCW(self):
        """Rotate counter-clockwise by 90 deg."""
        return self.T[::-1, :]

    @property
    def _funcs(self):
        """Return an index of possible callable functions in other modules, caching result if not alreadty there.

        Look in Stoner.Image.imagefuncs, scipy.ndimage.* an d scikit.* for functions. We assume that each function
        takes a first argument that is an ndarray of image data, so with __getattrr__ and _func_generator we
        can make a bound method through duck typing."""
        if self._func_proxy is None:  # Buyild the cache
            func_proxy = (
                regexpDict()
            )  # Cache is a regular expression dictionary - keys matched directly and then by regular expression

            # Get the Stoner.Image.imagefuncs mopdule first
            from Stoner.Image import imagefuncs

            for d in dir(imagefuncs):
                if not d.startswith("_"):
                    func = getattr(imagefuncs, d)
                    if callable(func) and func.__module__ == imagefuncs.__name__:
                        name = "{}__{}".format(func.__module__, d).replace(".", "__")
                        func_proxy[name] = func

            # Get the Stoner.Image.util mopdule next
            from Stoner.Image import util as SIutil

            for d in dir(SIutil):
                if not d.startswith("_"):
                    func = getattr(SIutil, d)
                    if callable(func) and func.__module__ == SIutil.__name__:
                        name = "{}__{}".format(func.__module__, d).replace(".", "__")
                        func_proxy[name] = func

            # Add scipy.ndimage functions
            import scipy.ndimage as ndi

            _sp_mods = [ndi.interpolation, ndi.filters, ndi.measurements, ndi.morphology, ndi.fourier]
            for mod in _sp_mods:
                for d in dir(mod):
                    if not d.startswith("_"):
                        func = getattr(mod, d)
                        if callable(func) and func.__module__ == mod.__name__:
                            func.transpose = True
                            name = "{}__{}".format(func.__module__, d).replace(".", "__")
                            func_proxy[name] = func
            # Add the scikit images modules
            _ski_modules = [
                color,
                exposure,
                feature,
                io,
                measure,
                filters,
                filters.rank,
                graph,
                util,
                restoration,
                morphology,
                segmentation,
                transform,
                viewer,
            ]
            for mod in _ski_modules:
                for d in dir(mod):
                    if not d.startswith("_"):
                        func = getattr(mod, d)
                        if callable(func):
                            name = "{}__{}".format(func.__module__, d).replace(".", "__")
                            func_proxy[name] = func
            self._func_proxy = func_proxy  # Store the cache
        return self._func_proxy

    # ==============================================================
    # function generator
    # ==============================================================
    def __dir__(self):
        """Implement code for dir()"""
        proxy = set(list(self._funcs.keys()))
        parent = set(dir(super(ImageArray, self)))
        mine = set(dir(ImageArray))
        return sorted(list(proxy | parent | mine))

    def __getattr__(self, name):
        """Magic attribute access method.

        Tries first to get the attribute via a superclass call, if this fails
        checks for some well known attribute names and supplies missing defaults.

        To handle magic calls into other modules, we have a regular expression dicitonary
        that stores and index of callables by full name where the . are changed to __
        If we can get a match to __<name> then we get that callable from our index.

        The callable is then passed to self._func_generator for wrapping into a
        'on the fly' method of this class.

        TODO:
            An alternative nested attribute system could be something like
            http://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-objects
            might be cool sometime.
        """
        ret = None
        try:
            ret = getattr(super(ImageArray, self), name)
        except AttributeError as e:
            # first check kermit funcs
            if name.startswith("_") or name in ["debug"]:
                if name == "_hardmask":
                    ret = False
            elif ".*__{}$".format(name) in self._funcs:
                ret = self._funcs[".*__{}$".format(name)]
                ret = self._func_generator(ret)
            if ret is None:
                raise AttributeError("No attribute found of name {}".format(name))
        return ret

    def _func_generator(self, workingfunc):
        """Used by __getattr__ to wrap an arbitary callbable to make it a bound method of this class.

        Args:
            workingfunc (callable): The callable object to be wrapped.

        Returns:
            (function): A function with enclosure that holds additional information about this object.

        The function returned from here will call workingfunc with the first argument being a clone of this
        ImageArray. If the meothd returns an ndarray, it is wrapped back to our own class and the metadata dictionary
        is updated. If the function returns a :py:class:`Stoner.Data` object then this is also updated with our metadata.

        This method also updates the name and documentation strings for the wrapper to match the wrapped function -
        thus ensuring that Spyder's help window can generate useful information.

        """

        @wraps(workingfunc)
        def gen_func(*args, **kwargs):
            transpose = getattr(workingfunc, "transpose", False)
            if transpose:
                change = self.clone.T
            else:
                change = self.clone
            r = workingfunc(change, *args, **kwargs)  # send copy of self as the first arg
            if isinstance(r, Data):
                pass  # Data return is ok
            elif isinstance(r, np.ndarray) and np.prod(r.shape) == np.max(r.shape):  # 1D Array
                r = Data(r)
                r.metadata = self.metadata.copy()
                r.column_headers[0] = workingfunc.__name__
            elif isinstance(r, np.ndarray):  # make sure we return a ImageArray
                if transpose:
                    r = r.view(type=self.__class__).T
                else:
                    r = r.view(type=self.__class__)
                sm = self.metadata.copy()  # Copy the currenty metadata
                sm.update(r.metadata)  # merge in any new metadata from the call
                r.metadata = sm  # and put the returned metadata as the merged data
            # NB we might not be returning an ndarray at all here !
            return r

        return fix_signature(gen_func, workingfunc)

    @property
    def draw(self):
        """DrawProxy is an opbject for accessing the skimage draw sub module."""
        return DrawProxy(self)

    # ==============================================================================
    # OTHER SPECIAL METHODS
    # ==============================================================================

    def __setattr__(self, name, value):
        """Set an attribute on the object."""
        super(ImageArray, self).__setattr__(name, value)
        # add attribute to those for copying in array_finalize. use value as
        # defualt.
        circ = ["_optinfo"]  # circular references
        proxy = ["_funcs"]  # can be reloaded for cloned arrays
        if name in circ + proxy:
            # Ignore these in clone
            pass
        else:
            self._optinfo.update({name: value})

    def __getitem__(self, index):
        """Patch indexing of strings to metadata."""
        if getattr(self, "debug", False):
            curframe = inspect.currentframe()
            calframe = inspect.getouterframes(curframe, 2)
        if isinstance(index, ImageFile) and index.image.dtype == bool:
            index = index.image
        if isinstance(index, string_types):
            return self.metadata[index]
        else:
            return super(ImageArray, self).__getitem__(index)

    def __setitem__(self, index, value):
        """Patch string index through to metadata."""
        if isinstance(index, ImageFile) and index.dtype == bool:
            index = index.image
        if isinstance(index, string_types):
            self.metadata[index] = value
        else:
            super(ImageArray, self).__setitem__(index, value)

    def __delitem__(self, index):
        """Patch indexing of strings to metadata."""
        if isinstance(index, string_types):
            del self.metadata[index]
        else:
            super(ImageArray, self).__delitem__(index)

    ############################################################################################################
    ############### Custom Methods for ImageArray###############################################################

    def crop(self, *args, **kargs):
        """Crop the image.

        This is essentially like taking a view onto the array
        but uses image x,y coords (x,y --> col,row)
        Returns a view according to the coords given. If box is None it will
        allow the user to select a rectangle. If a tuple is given with None
        included then max extent is used for that coord (analagous to slice).
        If copy then return a copy of self with the cropped image.

        Args:
            box(tuple) or 4 separate args or None:
                (xmin,xmax,ymin,ymax)
                If None image will be shown and user will be asked to select
                a box (bit experimental)

        Keyword Arguments:
            copy(bool):
                If True return a copy of ImageFile with the cropped image
        Returns:
            (ImageArray):
                view or copy of array asked for

        Example:
            a=ImageFile(np.arange(12).reshape(3,4))

            a.crop(1,3,None,None)
        """
        box = self._box(*args, **kargs)
        ret = self[box]
        if "copy" in kargs.keys() and kargs["copy"]:
            ret = ret.clone
        return ret

    def dtype_limits(self, clip_negative=True):
        """Return intensity limits, i.e. (min, max) tuple, of the image's dtype.

        Args:
            image(ndarray):
                Input image.
            clip_negative(bool):
                If True, clip the negative range (i.e. return 0 for min intensity)
                even if the image dtype allows negative values.

        Returns:
            (imin, imax : tuple): Lower and upper intensity limits.
        """
        if clip_negative is None:
            clip_negative = True
        imin, imax = dtype_range[self.dtype.type]
        if clip_negative:
            imin = 0
        return imin, imax

    def asfloat(self, normalise=True, clip=False, clip_negative=False):
        """Return the image converted to floating point type.

        If currently an int type and normalise then floats will be normalised
        to the maximum allowed value of the int type.
        If currently a float type then no change occurs.
        If clip then clip values outside the range -1,1
        If clip_negative then further clip values to range 0,1

        Keyword Arguments:
            normalise(bool):
                normalise the image to the max value of current int type
            clip_negative(bool):
                clip negative intensity to 0
        """
        if self.dtype.kind == "f":
            ret = self
        else:
            ret = convert(self, dtype=np.float64, normalise=normalise)  # preserve metadata
            ret = ImageArray(ret)
            c = self.clone  # copy formatting and apply to new array
            for k, v in c._optinfo.items():
                setattr(ret, k, v)
        if clip or clip_negative:
            ret = ret.clip_intensity(clip_negative=clip_negative)
        return ret

    def clip_intensity(self, clip_negative=False, limits=None):
        """Clip intensity outside the range -1,1 or 0,1

        Keyword ArgumentsL
            clip_negative(bool):
                if True clip to range 0,1 else range -1,1
            limits (low,high): Clip the intensity between low and high rather than zero and 1.

        Ensure data range is -1 to 1 or 0 to 1 if clip_negative is True.

        """
        if limits is None:
            dl = self.dtype_limits(clip_negative=clip_negative)
        else:
            dl = list(limits)
        np.clip(self, dl[0], dl[1], out=self)

    def asint(self, dtype=np.uint16):
        """convert the image to unsigned integer format.

        May raise warnings about loss of precision.
        """
        if self.dtype.kind == "f" and (np.max(self) > 1 or np.min(self) < -1):
            self = self.normalise()
        ret = convert(self, dtype)
        return ret

    def save(self, filename=None, **kargs):
        """Saves the image into the file 'filename'.

        Metadata will be preserved in .png and .tif format.

        fmt can be 'png', 'npy', 'tif', 'tiff'  or a list of more than one of those.
        tif is recommended since metadata is lost in .npy format but data is
        converted to integer format for png so that definition cannot be
        saved.

        Args:
            filename (string, bool or None): Filename to save data as, if this is
                None then the current filename for the object is used
                If this is not set, then then a file dialog is used. If
                filename is False then a file dialog is forced.
        Keyword args:
            fmt (string or list): format to save data as. 'tif', 'png' or 'npy'
                or a list of them. If not included will guess from filename.

            forcetype (bool): integer data will be converted to np.float32 type
                for saving. if forcetype then preserve and save as int type (will
                                                                             be unsigned).

        Since Stoner.Image is meant to be a general 2d array often with negative
        and floating point data this poses a problem for saving images. Images
        are naturally saved as 8 or more bit unsigned integer values representing colour.
        The only obvious way to save an image and preserve negative data
        is to save as a float32 tif. This has the advantage over the npy
        data type which cannot be opened by external programs and will not
        save metadata.
        """
        # Standard filename block
        if filename is None:
            filename = self.filename
        if filename is None or (isinstance(filename, bool) and not filename):
            # now go and ask for one
            filename = self.__file_dialog("w")

        def_fmt = os.path.splitext(filename)[1][1:]  # Get a default format from the filename
        if def_fmt not in self.fmts:  # Default to png if nothing else
            def_fmt = "png"
        fmt = kargs.pop("fmt", [def_fmt])

        if not isinstance(fmt, list):
            fmt = [fmt]
        if set(fmt) & set(self.fmts) == set([]):
            raise ValueError("fmt must be {}".format(",".join(self.fmts)))
        fmt = ["tiff" if f == "tif" else f for f in fmt]
        self.filename = filename
        for fm in fmt:
            saver = getattr(self, "save_{}".format(fm), "save_tif")
            if fm == "tiff":
                forcetype = kargs.pop("forcetype", False)
                saver(filename, forcetype)
            else:
                saver(filename)

    def save_png(self, filename):
        """Save the ImageArray with metadata in a png file.
        This can only save as 8bit unsigned integer so there is likely
        to be a loss of precision on floating point data"""
        pngname = os.path.splitext(filename)[0] + ".png"
        meta = PngImagePlugin.PngInfo()
        info = self.metadata.export_all()
        info = [(i.split("=")[0], "=".join(i.split("=")[1:])) for i in info]
        for k, v in info:
            meta.add_text(k, v)
        s = (self - self.min()) * 256 / (self.max() - self.min())
        im = Image.fromarray(s.astype("uint8"), mode="L")
        im.save(pngname, pnginfo=meta)

    def save_npy(self, filename):
        """Save the ImageArray as a numpy array."""
        npyname = os.path.splitext(filename)[0] + ".npy"
        np.save(npyname, np.array(self))

    def save_tiff(self, filename, forcetype=False):
        """Save the ImageArray as a tiff image with metadata
        PIL can save in modes "L" (8bit unsigned int), "I" (32bit signed int),
        or "F" (32bit signed float). In general max info is preserved for "F"
        type so if forcetype is not specified then this is the default. For
        boolean type data mode "L" will suffice and this is chosen in all cases.
        The type name is added as a string to the metadata before saving.

        Keyword Args:
            forcetype(bool): if forcetype then preserve data type as best as
                possible on save.
                Otherwise integer data will be converted to np.float32 type
                for saving. (bool will remain as int since there's no danger of
                loss of information)
        """
        from PIL.TiffImagePlugin import ImageFileDirectory_v2
        import json

        dtype = np.dtype(self.dtype).name  # string representation of dtype we can save
        self["ImageArray.dtype"] = dtype  # add the dtype to the metadata for saving.
        if forcetype:  # PIL supports uint8, int32 and float32, try to find the best match
            if self.dtype == np.uint8 or self.dtype.kind == "b":  # uint8 or boolean
                im = Image.fromarray(self, mode="L")
            elif self.dtype.kind in ["i", "u"]:
                im = Image.fromarray(self, mode="I")
            else:  # default to float32
                im = Image.fromarray(self.astype(np.float32), mode="F")
        else:
            if self.dtype.kind == "b":  # boolean we're not going to lose data by saving as unsigned int
                im = Image.fromarray(self, mode="L")
            else:  # try to convert everything else to float32 which can has maximum preservation of info
                im = Image.fromarray(self.astype(np.float32), mode="F")
        ifd = ImageFileDirectory_v2()
        ifd[270] = json.dumps(self.metadata.export_all())
        ext = os.path.splitext(filename)[1]
        if ext in [".tif", ".tiff"]:  # ensure extension is preserved in save
            pass
        else:  # default to tiff
            ext = ".tiff"
        tiffname = os.path.splitext(filename)[0] + ext
        im.save(tiffname, tiffinfo=ifd)

    def __file_dialog(self, mode):
        """Creates a file dialog box for loading or saving ~b ImageFile objects.

        Args:
            mode(string): The mode of the file operation  'r' or 'w'

        Returns:
            A filename to be used for the file operation.
        """
        patterns = ("png", "*.png")

        if self.filename is not None:
            filename = os.path.basename(self.filename)
            dirname = os.path.dirname(self.filename)
        else:
            filename = ""
            dirname = ""
        if "r" in mode:
            mode = "file"
        elif "w" in mode:
            mode = "save"
        else:
            mode = "directory"
        dlg = get_filedialog(what=mode, initialdir=dirname, initialfile=filename, filetypes=patterns)
        if len(dlg) != 0:
            self.filename = dlg
            return self.filename
        else:
            return None

    ############################################################################################################
    ############## Depricated Methods ##########################################################################

    def box(self, *args, **kargs):
        """Alias for :py:meth:`ImageArray.crop`"""
        warnings.warn("The box method was replaced by crop and will raise and error in future versions.")
        return self.crop(*args, **kargs)

    def crop_image(self, *args, **kargs):
        """Back compatability alias for :py:meth:`ImageArray.crop`"""
        warnings.warn("The crop_image method was replaced by crop and will raise and error in future versions.")
        return self.crop(*args, **kargs)

    def convert_float(self, clip_neg=True):
        """Deproicated compatability. :py:meth:`ImageArray.asfloat` preferred"""
        warnings.warn("The convert_float method was replaced by asfloat and will raise and error in future versions.")
        self.asfloat(normalise=False, clip_negative=clip_neg)

    def convert_int(self):
        """Depricated compatability meothd. :py:meth:`ImageArray.asint` preferred"""
        warnings.warn("The convert_int method was replaced by asint and will raise and error in future versions.")
        self.asint()


class ImageFile(metadataObject):

    """An Image file type that is analagous to DataFile.

    This contains metadata and an image attribute which
    is an ImageArray type which subclasses numpy ndarray and adds lots of extra
    image specific processing functions.

    The ImageFile owned attribute is image. All other calls including metadata
    are passed through to ImageArray (so no need to inherit from metadataObject).

    Almost all calls to ImageFile are passed through to the underlying ImageArray
    logic and ImageArray can be used as a standalone class.
    However because ImageArray subclasses an ndarray it is not possible to enter
    it in place. All attributes return an array instance which needs to be reassigned.
    ImageFile owns image and so can change in place.
    The penalty is that numpy ufuncs don't return ImageFile type

    so can do::

        imfile.asfloat() #imagefile.image is updated to float type however need to do:
        imfile.image = np.abs(imfile.image)

    whereas for imarray need to do::

            imarray = imagearray.asfloat()

    but::

            np.abs(imarray) #returns ImageArray type
    """

    _image = None
    _protected_attrs = ["_fromstack"]  # these won't be passed through to self.image attrs

    def __init__(self, *args, **kargs):
        """Mostly a pass through to ImageArray constructor.

        Local attribute is image. All other attributes and calls are passed
        through to image attribute.

        There is one special case of creating an ImageFile from a :py:class:`Stoner.Core.DataFile`. In this case the
        the DataFile is assummed to contain (x,y,z) data that should be converted to a map of
        z on a regular grid of x,y. The columns for the x,y,z data can be taken from the DataFile's
        :py:attr:`Stoner.Core.DataFile.setas` attribute or overridden by providing xcol, ycol and zcol keyword arguments.
        A further *shape* keyword can spewcify the shape as a tuple or "unique" to use the unique values of x and y or if
        omitted asquare grid will be interpolated.

        """
        super(ImageFile, self).__init__(*args, **kargs)
        if len(args) == 0:
            self._image = ImageArray()
        elif len(args) > 0 and isinstance(args[0], string_types):
            self._image = ImageArray(*args, **kargs)
        elif len(args) > 0 and isinstance(args[0], ImageFile):  # Fixing type
            self._image = args[0].image
        elif len(args) > 0 and isinstance(args[0], np.ndarray):  # Fixing type
            self._image = ImageArray(*args, **kargs)
        elif len(args) > 0 and isinstance(
            args[0], DataFile
        ):  # Support initing from a DataFile that defines x,y,z coordinates
            self._init_from_datafile(*args, **kargs)
        self._fromstack = kargs.pop("_fromstack", False)  # for use by ImageStack

    #####################################################################################################################################
    ############################# Properties #### #######################################################################################

    @property
    def _repr_png_(self):
        if get_option("short_repr") or get_option("short_img_repr"):
            raise AttrbuteError("Suppressed graphical representation")
        else:
            return self._repr_png_private_

    @property
    def clone(self):
        """Make a copy of this ImageFile."""
        new = self.__class__(self.image.clone)
        for attr in self.__dict__:
            if callable(getattr(self, attr)) or attr in ["image", "metadata"]:
                continue
            try:
                setattr(new, attr, deepcopy(getattr(self, attr)))
            except NotImplementedError:  # Deepcopying failed, so just copy a reference instead
                setattr(new, attr, getattr(self, attr))
        return new

    @property
    def data(self):
        """alias for image[:]. Equivalence to Stoner.data behaviour"""
        return self.image

    @data.setter
    def data(self, value):
        """Simple minded pass through."""
        self.image = value

    @property
    def CW(self):
        """Rotate clockwise by 90 deg."""
        ret = self.clone
        ret.image = ret.image.CW
        return ret

    @property
    def CCW(self):
        """Rotate counter-clockwise by 90 deg."""
        ret = self.clone
        ret.image = ret.image.CCW
        return ret

    @property
    def image(self):
        """Access the image data."""
        return self._image

    @image.setter
    def image(self, v):
        """Ensure stored image is always an ImageArray."""
        filename = self.filename
        # ensure setting image goes into the same memory block if from stack
        if (
            hasattr(self, "_fromstack")
            and self._fromstack
            and self._image.shape == v.shape
            and self._image.dtype == v.dtype
        ):
            self._image[:] = v
        else:
            self._image = ImageArray(v)
        self.filename = filename

    @property
    def filename(self):
        """Pass through to image attribute."""
        if self._image is not None:
            return self.image.filename
        else:
            return ""

    @filename.setter
    def filename(self, value):
        """Pass through to image attribute."""
        if self._image is None:
            self._image = ImageArray()
        self.image.filename = value

    @property
    def mask(self):
        """Get the mask of the underlying IamgeArray."""
        return MaskProxy(self)

    @mask.setter
    def mask(self, value):
        """Set the underlying ImageArray's mask."""
        if isinstance(value, ImageFile):
            value = value.image
        if isinstance(value, MaskProxy):
            value = value._mask
        self.image.mask = value

    @property
    def metadata(self):
        """Intercept metadata and redirect to image.metadata."""
        return self.image.metadata

    @metadata.setter
    def metadata(self, value):
        self.image.metadata = value

    #####################################################################################################################################
    ############################# Special methods #######################################################################################

    def __dir__(self):
        """Implement code for dir()"""
        proxy = set(dir(self.image))
        parent = set(dir(super(ImageFile, self)))
        mine = set(dir(ImageFile))
        return sorted(list(proxy | parent | mine))

    def __getitem__(self, n):
        """A pass through to ImageArray."""
        try:
            return self.image.__getitem__(n)
        except KeyError:
            if n not in self.metadata and n in self._image.metadata:
                self.metadata[n] = self._image.metadata[n]
            return self.metadata.__getitem__(n)

    def __setitem__(self, n, v):
        """A Pass through to ImageArray."""
        if isinstance(n, string_types):
            self.metadata.__setitem__(n, v)
        else:
            self.image.__setitem__(n, v)

    def __delitem__(self, n):
        """A Pass through to ImageArray."""
        try:
            self.image.__delitem__(n)
        except KeyError:
            self.metadata.__delitem__(n)

    def __getattr__(self, n):
        """"Handles attriobute access."""
        try:
            ret = super(ImageFile, self).__getattr__(n)
        except AttributeError:
            ret = getattr(self.image, n)
            if callable(ret):  # we have a method
                ret = self._func_generator(ret)  # modiy so that we can change image in place
        return ret

    def __setattr__(self, n, v):
        """Handles setting attributes."""
        if not hasattr(self, n) and n not in ImageFile._protected_attrs:
            setattr(self._image, n, v)
        else:
            super(ImageFile, self).__setattr__(n, v)

    def __add__(self, other):
        """Implement the subtract operator"""
        result = self.clone
        result = self.__add_core__(result, other)
        return result

    def __iadd__(self, other):
        """Implement the inplace subtract operator"""
        result = self
        result = self.__add_core__(result, other)
        return result

    def __add_core__(self, result, other):
        """Actually do result=result-other."""
        if isinstance(other, result.__class__) and result.shape == other.shape:
            result.image += other.image
        elif isinstance(other, np.ndarray) and other.shape == result.shape:
            result.image += other
        elif isinstance(other, (int, float)):
            result.image += other
        else:
            return NotImplemented
        return result

    if python_v3:

        def __truediv__(self, other):
            """Implement the divide operator"""
            result = self.clone
            result = self.__div_core__(result, other)
            return result

        def __itruediv__(self, other):
            """Implement the inplace divide operator"""
            result = self
            result = self.__div_core__(result, other)
            return result

    else:

        def __div__(self, other):
            """Implement the divide operator"""
            result = self.clone
            result = self.__div_core__(result, other)
            return result

        def __idiv__(self, other):
            """Implement the inplace divide operator"""
            result = self
            result = self.__div_core__(result, other)
            return result

    def __div_core__(self, result, other):
        """Actually do result=result/other."""
        # Cheat and pass through to ImageArray

        if isinstance(other, ImageFile):
            other = other.image

        result.image = result.image / other
        return result

    def __sub__(self, other):
        """Implement the subtract operator"""
        result = self.clone
        result = self.__sub_core__(result, other)
        return result

    def __isub__(self, other):
        """Implement the inplace subtract operator"""
        result = self
        result = self.__sub_core__(result, other)
        return result

    def __sub_core__(self, result, other):
        """Actually do result=result-other."""
        if isinstance(other, result.__class__) and result.shape == other.shape:
            result.image -= other.image
        elif isinstance(other, np.ndarray) and other.shape == result.shape:
            result.image -= other
        elif isinstance(other, (int, float)):
            result.image -= other
        else:
            return NotImplemented
        return result

    def __neg__(self):
        """Intelliegent negate function that handles unsigned integers."""
        ret = self.clone
        if self._image.dtype.kind == "u":
            l, h = dtype_range[self._image.dtype]
            ret.image = h - self.image
        else:
            ret.image = -self.image
        return ret

    def __invert__(self):
        """Equivalent to clockwise rotation"""
        return self.CW

    def __floordiv__(self, other):
        """Calculate and XMCD ratio on the images."""
        if not isinstance(other, ImageFile):
            return NotImplemented
        ret = self.clone
        ret.image = (self.image - other.image) / (self.image + other.image)
        return ret

    #####################################################################################################################################
    ############################# Private methods #######################################################################################

    def _init_from_datafile(self, *args, **kargs):
        """Initialise ImageFile from DataFile defining x,y,z co-ordinates.

        Args:
            args[0] (DataFile): A :py:class:`Stoner.Core.DataFile` instance that defines x,y,z co-ordinates or has columns specified in keywords.

        Keyword Args:
            xcol (column index): Column in the DataFile that has the x-co-ordinate
            ycol (column index): Column in the data file that defines the y-cordinate
            zcol (column index): Column in the datafile that defines the intensity
        """
        data = Data(args[0])
        shape = kargs.pop("shape", "unique")

        _ = data._col_args(**kargs)
        data.setas(x=_.xcol, y=_.ycol, z=_.zcol)
        if isinstance(shape, string_types) and shape == "unique":
            shape = (len(np.unique(data.x)), len(np.unique(data.y)))
        elif istuple(shape, int_types, int_types):
            pass
        else:
            shape = None
        X, Y, Z = data.griddata(_.xcol, _.ycol, _.zcol, shape=shape)
        self.image = Z
        self.metadata = deepcopy(data.metadata)
        self["x_vector"] = np.unique(X)
        self["y_vector"] = np.unique(Y)

    def _func_generator(self, workingfunc):
        """ImageFile generator.

        Note:
            The wrapped functions take additional keyword arguments that are stripped off from the call.

            _box(:py:meth:`Stoner.ImageArray.crop` arguments): Crops the image first before calling the parent method.
            _ (bool, None): Controls whether a :py:class:`ImageArray` return will be substituted for the current :py:class:`ImageArray`.

            * True: - all ImageArray return types are substituted.
            * False (default) - Imagearray return types are substituted if they are the same size as the original
            * None - A copy of the current object is taken and the returned ImageArray provides the data.
        """

        @wraps(workingfunc)
        def gen_func(*args, **kargs):

            box = kargs.pop("_box", None)
            if len(args) > 0:
                args = list(args)
                for ix, a in enumerate(args):
                    if isinstance(a, ImageFile):
                        args[ix] = a.image[self.image._box(box)]
            if (
                workingfunc.__name__ == "crop" and "_" not in kargs.keys()
            ):  # special case for common function crop which will change the array shape
                force = True
            else:
                force = kargs.pop("_", False)
            r = workingfunc(*args, **kargs)
            if (
                force is not None
                and isinstance(r, ImageArray)
                and (force or r.shape == self.image[self.image._box(box)].shape)
            ):
                # Enure that we've captured any metadata added inside the working function
                self.metadata.update(r.metadata)
                # Now swap the iamge in, but keep the metadata
                r.metadata = self.metadata
                filename = self.filename
                self.image = r
                self.filename = filename

                return self
            elif force is None:
                ret = self.clone
                ret.metadata.update(r.metadata)
                # Now swap the iamge in, but keep the metadata
                r.metadata = ret.metadata
                filename = ret.filename
                ret.image = r
                ret.filename = filename
                return ret
            else:
                return r

        return fix_signature(gen_func, workingfunc)

    def __repr__(self):
        return "{}({}) of shape {} ({}) and {} items of metadata".format(
            self.filename, type(self), self.shape, self.image.dtype, len(self.metadata)
        )

    def _repr_png_private_(self):
        """Provide a display function for iPython/Jupyter."""
        fig = self.image.imshow()
        plt.title(self.filename)
        data = StreamIO()
        fig.savefig(data, format="png")
        plt.close(fig)
        data.seek(0)
        ret = data.read()
        data.close()
        return ret

    def save(self, *args, **kwargs):
        """Saves the image into the file 'filename'.

        Metadata will be preserved in .png and .tif format.

        fmt can be 'png', 'npy', 'tif', 'tiff'  or a list of more than one of those.
        tif is recommended since metadata is lost in .npy format but data is
        converted to integer format for png so that definition cannot be
        saved.

        Args:
            filename (string, bool or None): Filename to save data as, if this is
                None then the current filename for the object is used
                If this is not set, then then a file dialog is used. If
                filename is False then a file dialog is forced.
        Keyword args:
            fmt (string or list): format to save data as. 'tif', 'png' or 'npy'
            or a list of them. If not included will guess from filename.
        """
        # catch before metadataObject tries to take over.
        self.image.save(*args, **kwargs)


class DrawProxy(object):
    """Provides a wrapper around scikit-image.draw to allow easy drawing of objects onto images."""

    def __init__(self, *args, **kargs):
        """Grab the parent image from the constructor."""
        self.img = args[0]

    def __getattr__(self, name):
        """Retiurn a callable function that will carry out the draw operation requested."""
        func = getattr(draw, name)

        @wraps(func)
        def _proxy(*args, **kargs):
            value = kargs.pop("value", np.ones(1, dtype=self.img.dtype)[0])
            coords = func(*args, **kargs)
            self.img[coords] = value
            return self.img

        return fix_signature(_proxy, func)

    def __dir__(self):
        """Pass through to the dir of skimage.draw."""
        return draw.__dir__()

    def annulus(self, r, c, radius1, radius2, shape=None, value=1.0):
        """Use a combination of two circles to draw and annulus.

        Args:
            r,c (float): Centre co-ordinates
            radius1,radius2 (float): Inner and outer radius.

        Keyword Arguments:
            shape (2-tuple, None): Confine the co-ordinates to staywith shape
            value (float): value to draw with
        Returns:
            A copy of the image with the annulus drawn on it.

        Notes:
            If radius2<radius1 then the sense of the whole shape is inverted
            so that the annulus is left clear and the filed is filled.
        """
        if shape is None:
            shape = self.img.shape
        invert = radius2 < radius1
        if invert:
            buf = np.ones(shape)
            fill = 0.0
            bg = 1.0
        else:
            buf = np.zeros(shape)
            fill = 1.0
            bg = 2.0
        radius1, radius2 = min(radius1, radius2), max(radius1, radius2)
        rr, cc = draw.circle(r, c, radius2, shape=shape)
        buf[rr, cc] = fill
        rr, cc = draw.circle(r, c, radius1, shape=shape)
        buf[rr, cc] = bg
        self.img = np.where(buf == 1, value, self.img)
        return self.img

    def rectangle(self, r, c, w, h, angle=0.0, shape=None, value=1.0):
        """Draw a rectangle on an image.

        Args:
            r,c (float): Centre co-ordinates
            w,h (float): Lengths of the two sides of the rectangle

        Keyword Arguments:
            angle (float): Angle to rotate the rectangle about
            shape (2-tuple or None): Confine the co-ordinates to this shape.
            value (float): The value to draw with.

        Returns:
            A copy of the current image with a rectangle drawn on.
        """
        if shape is None:
            shape = self.img.shape

        x1 = r - h / 2
        x2 = r + h / 2
        y1 = c - w / 2
        y2 = c + w / 2
        co_ords = np.array([[x1, y1], [x2, y1], [x2, y2], [x1, y2]])
        if angle != 0:
            centre = np.array([r, c])
            c, s, m = np.cos, np.sin, np.matmul
            r = np.array([[c(angle), -s(angle)], [s(angle), c(angle)]])
            co_ords = np.array([centre + m(r, xy - centre) for xy in co_ords])
        rr, cc = draw.polygon(co_ords[:, 0], co_ords[:, 1], shape=shape)
        self.img[rr, cc] = value
        return self.img

    def square(self, r, c, w, angle=0.0, shape=None, value=1.0):
        """Draw a square on an image.

        Args:
            r,c (float): Centre co-ordinates
            w (float): Length of the side of the square

        Keyword Arguments:
            angle (float): Angle to rotate the rectangle about
            shape (2-tuple or None): Confine the co-ordinates to this shape.
            value (float): The value to draw with.

        Returns:
            A copy of the current image with a rectangle drawn on.
        """
        return self.rectangle(r, c, w, w, angle=angle, shape=shape, value=value)


class MaskProxy(object):
    @property
    def _IA(self):
        return self._IF.image

    @property
    def _mask(self):
        self._IA.mask = np.ma.getmaskarray(self._IA)
        return self._IA.mask

    @property
    def data(self):
        return self[:]

    @property
    def image(self):
        return self[:]

    @property
    def draw(self):
        return DrawProxy(self._mask)

    def __init__(self, *args, **kargs):
        """Keep track of the underlying objects."""
        self._IF = args[0]

    def __getitem__(self, index):
        """Proxy through to mask index."""
        return self._mask.__getitem__(index)

    def __setitem__(self, index, value):
        """Proxy through to underlying mask."""
        self._IA.mask.__setitem__(index, value)

    def __delitem__(self, index):
        """Proxy through to underyling mask."""
        self._IA.mask.__delitem__(index)

    def __getattr__(self, name):
        """Checks name against self._IA._funcs and constructs a method to edit the mask as an image."""
        if hasattr(self._IA.mask, name):
            return getattr(self._IA.mask, name)
        if not ".*__{}$".format(name) in self._IA._funcs:
            raise AttributeError("{} not a callable mask method.".format(name))
        func = self._IA._funcs[".*__{}$".format(name)]

        @wraps(func)
        def _proxy_call(*args, **kargs):
            r = func(self._mask.astype(int), *args, **kargs)
            if isinstance(r, np.ndarray) and r.shape == self._IA.shape:
                self._IA.mask = r
            return r

        _proxy_call.__doc__ = func.__doc__
        _proxy_call.__name__ = func.__name__
        return fix_signature(_proxy_call, func)

    def __rep__(self):
        return repr(self._mask)

    def __str__(self):
        return repr(self._mask)

    def __neg__(self):
        return -self._mask

    def _repr_png_(self):
        """Provide a display function for iPython/Jupyter."""
        fig = self._IA._funcs[".*imshow"](self._mask.astype(int))
        data = StreamIO()
        fig.savefig(data, format="png")
        plt.close(fig)
        data.seek(0)
        ret = data.read()
        data.close()
        return ret

    def clear(self):
        """Clear a mask."""
        self._IA.mask = np.zeros_like(self._IA)

    def invert(self):
        """Invert the mask."""
        self._IA.mask = ~self._IA.mask



# -*- coding: utf-8 -*-
"""
test_Core.py
Created on Tue Jan 07 22:05:55 2014

@author: phygbu
"""

import unittest
import sys
import os, os.path as path
import numpy as np
import re
from numpy import any,all,sqrt,nan
from collections import MutableMapping,OrderedDict

pth=path.dirname(__file__)
pth=path.realpath(path.join(pth,"../../"))
sys.path.insert(0,pth)
from Stoner import Data,__home__
from Stoner.Core import typeHintedDict,metadataObject
import Stoner.compat

def extra_get_filedialog(what="file", **opts):
    return path.join(path.dirname(__file__),"CoreTest.dat")

def mask_func(r):
    return np.abs(r.q)<0.25*np.pi


Stoner.Core.get_filedialog=extra_get_filedialog #Monkey patch to test the file dialog

class Datatest(unittest.TestCase):

    """Path to sample Data File"""
    datadir=path.join(pth,"sample-data")

    def setUp(self):
        self.d=Data(path.join(path.dirname(__file__),"CoreTest.dat"),setas="xy")
        self.d2=Data(path.join(__home__,"..","sample-data","TDI_Format_RT.txt"))
        self.d3=Data(path.join(__home__,"..","sample-data","New-XRay-Data.dql"))
        self.d4=Data(path.join(__home__,"..","sample-data","Cu_resistivity_vs_T.txt"))

    def test_repr(self):
        header="""==============  ======  ======
TDI Format 1.5  X-Data  Y-Data
    index       0 (x)   1 (y)
==============  ======  ======"""
        repr_header="""=============================  ========  ========
TDI Format 1.5                   X-Data    Y-Data
index                             0 (x)     1 (y)
=============================  ========  ========"""
        self.assertEqual(self.d.header,header,"Header output changed.")
        self.assertEqual("\n".join(repr(self.d).split("\n")[:4]),repr_header,"Representation gave unexpcted header.")

    def test_base_class(self):
        m=metadataObject()
        m.update(self.d2.metadata)
        self.assertTrue(isinstance(m,MutableMapping),"metadataObject is not a MutableMapping")
        self.assertEqual(len(m),len(self.d2.metadata),"metadataObject length failure.")
        self.assertEqual(len(m[".*"]),len(m),"Failure to index by regexp.")
        for k in self.d2.metadata:
            self.assertEqual(m[k],self.d2.metadata[k],"Failure to have equal keys for {}".format(k))
            m[k]=k
            self.assertEqual(m[k],k,"Failure to set keys for {}".format(k))
            del m[k]
        self.assertEqual(len(m),0,"Failed to delete from metadataObject.")

    def test_constructor(self):
        """Constructor Tests"""
        d=Data()
        self.assertTrue(d.shape==(1,0),"Bare constructor failed")
        d=Data(self.d)
        self.assertTrue(np.all(d.data==self.d.data),"Constructor from DataFile failed")
        d=Data([np.ones(100),np.zeros(100)])
        self.assertTrue(d.shape==(100,2),"Constructor from iterable list of nd array failed")
        d=Data([np.ones(100),np.zeros(100)],["X","Y"])
        self.assertTrue(d.column_headers==["X","Y"],"Failed to set column headers in constructor: {}".format(d.column_headers))
        c=np.zeros(100)
        d=Data({"X-Data":c,"Y-Data":c,"Z-Data":c})
        self.assertEqual(d.shape,(100,3),"Construction from dictionary of columns failed.")
        d=Data(False)
        self.assertEqual(self.d,d,"Faked file-dialog test")

    def test_column(self):
        for i,c in enumerate(self.d.column_headers):
            # Column function checks
            self.assertTrue(all(self.d.data[:,i]==self.d.column(i)),"Failed to Access column {} by index".format(i))
            self.assertTrue(all(self.d.data[:,i]==self.d.column(c)),"Failed to Access column {} by name".format(i))
            self.assertTrue(all(self.d.data[:,i]==self.d.column(c[0:3])),"Failed to Access column {} by partial name".format(i))

        # Check that access by list of strings returns multpiple columns
        self.assertTrue(all(self.d.column(self.d.column_headers)==self.d.data),"Failed to access all columns by list of string indices")
        self.assertTrue(all(self.d.column([0,self.d.column_headers[1]])==self.d.data),"Failed to access all columns by mixed list of indices")

        # Check regular expression column access
        self.assertTrue(all(self.d.column(re.compile(r"^X\-"))[:,0]==self.d.column(0)),"Failed to access column by regular expression")
        # Check attribute column access
        self.assertTrue(all(self.d.X==self.d.column(0)),"Failed to access column by attribute name")

    def test_deltions(self):
        ch=["{}-Data".format(chr(x)) for x in range(65,91)]
        data=np.zeros((100,26))
        metadata=OrderedDict([("Key 1",True),("Key 2",12),("Key 3","Hellow world")])
        self.dd=Data(metadata)
        self.dd.data=data
        self.dd.column_headers=ch
        self.dd.setas="3.x3.y3.z"
        self.repr_string="""===========================  ========  =======  ========  =======  ========  ========
TDI Format 1.5                 D-Data   ....      H-Data   ....      Y-Data    Z-Data
index                           3 (x)              7 (y)                 24        25
===========================  ========  =======  ========  =======  ========  ========
Key 1{Boolean}= True                0                  0                  0         0
Key 2{I32}= 12                      0  ...             0  ...             0         0
Key 3{String}= Hellow world         0  ...             0  ...             0         0
Stoner.class{String}= Data          0  ...             0  ...             0         0
...                                 0  ...             0  ...             0         0"""
        self.assertEqual("\n".join(repr(self.dd).split("\n")[:9]),self.repr_string,"Representation with interesting columns failed.")
        del self.dd["Key 1"]
        self.assertEqual(len(self.dd.metadata),3,"Deletion of metadata failed.")
        del self.dd[20:30]
        self.assertEqual(self.dd.shape,(90,26),"Deleting rows directly failed.")
        self.dd.del_column("Q")
        self.assertEqual(self.dd.shape,(90,25),"Deleting rows directly failed.")
        self.dd%=3
        self.assertEqual(self.dd.shape,(90,24),"Deleting rows directly failed.")
        self.dd.setas="x..y..z"
        self.dd.del_column(self.dd.setas.not_set)
        self.assertEqual(self.dd.shape,(90,3),"Deleting rows directly failed.")
        self.dd.mask[50,1]=True
        self.dd.del_column()
        self.assertEqual(self.dd.shape,(90,2),"Deletion of masked rows failed.")

    def test_indexing(self):
        #Check all the indexing possibilities
        data=np.array(self.d.data)
        colname=self.d.column_headers[0]
        self.assertTrue(all(self.d.column(colname)==self.d[:,0]),"Failed direct indexing versus column method")
        self.assertTrue(all(self.d[:,0]==data[:,0]),"Failed direct idnexing versusus direct array index")
        self.assertTrue(all(self.d[:,[0,1]]==data),"Failed direct list indexing")
        self.assertTrue(all(self.d[::2,:]==data[::2]),"Failed slice indexing rows")
        self.assertTrue(all(self.d[colname]==data[:,0]),"Failed direct indexing by column name")
        self.assertTrue(all(self.d[:,colname]==data[:,0]),"Failed fallback indexing by column name")
        self.assertEqual(self.d[25,1],645.0,"Failed direct single cell index")
        self.assertEqual(self.d[25,"Y-Data"],645.0,"Failoed single cell index direct")
        self.assertEqual(self.d["Y-Data",25],645.0,"Failoed single cell fallback index order")
        self.d["X-Dat"]=[11,12,13,14,15]
        self.assertEqual(self.d["X-Dat",2],13,"Failed indexing of metadata lists with tuple")
        self.assertEqual(self.d["X-Dat"][2],13,"Failed indexing of metadata lists with double indices")
        d=Data(np.ones((10,10)))
        d[0,0]=5 #Index by tuple into data
        d["Column_1",0]=6 # Index by column name, row into data
        d[0,"Column_2"]=7 #Index by row, column name into data
        d["Column_3"]=[1,2,3,4] # Create a metadata
        d["Column_3",2]=2 # Index existing metadata via tuple
        d.metadata[0,5]=10
        d[0,5]=12 # Even if tuple, index metadata if already existing.
        self.assertTrue(np.all(d[0]==np.array([5,6,7,1,1,1,1,1,1,1])),"setitem on Data to index into Data.data failed.\n{}".format(d[0]))
        self.assertEqual(d.metadata["Column_3"],[1,2,2,4],"Tuple indexing into metadata Failed.")
        self.assertEqual(d.metadata[0,5],12,"Indexing of pre-existing metadta keys rather than Data./data failed.")
        self.assertEqual(d.metadata[1],[1, 2, 2, 4],"Indexing metadata by integer failed.")

    def test_len(self):
        # Check that length of the column is the same as length of the data
        self.assertEqual(len(Data()),0,"Empty DataFile not length zero")
        self.assertEqual(len(self.d.column(0)),len(self.d),"Column 0 length not equal to DataFile length")
        self.assertEqual(len(self.d),self.d.data.shape[0],"DataFile length not equal to data.shape[0]")
        # Check that self.column_headers returns the right length
        self.assertEqual(len(self.d.column_headers),self.d.data.shape[1],"Length of column_headers not equal to data.shape[1]")

    def test_attributes(self):
        """Test various atribute accesses,"""
        self.assertEqual(self.d.shape,(100,2),"shape attribute not correct.")
        self.assertTrue(all(self.d.T==self.d.data.T),"Transpose attribute not right")
        self.assertTrue(all(self.d.x==self.d.column(0)),"x attribute quick access not right.")
        self.assertTrue(all(self.d.y==self.d.column(1)),"y attribute not right.")
        self.assertTrue(all(self.d.q==np.arctan2(self.d.data[:,0],self.d.data[:,1])),"Calculated theta attribute not right.")
        self.assertTrue(all(sqrt(self.d[:,0]**2+self.d[:,1]**2)==self.d.r),"Calculated r attribute not right.")
        self.assertTrue(self.d2.records._[5]["Column 2"]==self.d2.data[5,2],"Records and as array attributes problem")

    def test_iterators(self):
        for i,c in enumerate(self.d.columns()):
            self.assertTrue(all(self.d.column(i)==c),"Iterating over DataFile.columns not the same as direct indexing column")
        for j,r in enumerate(self.d.rows()):
            self.assertTrue(all(self.d[j]==r),"Iteratinf over DataFile.rows not the same as indexed access")
        for k,r in enumerate(self.d):
            pass
        self.assertEqual(j,k,"Iterating over DataFile not the same as DataFile.rows")
        self.assertEqual(self.d.data.shape,(j+1,i+1),"Iterating over rows and columns not the same as data.shape")

    def test_metadata(self):
        self.d["Test"]="This is a test"
        self.d["Int"]=1
        self.d["Float"]=1.0
        keys,values=zip(*self.d.items())
        self.assertEqual(tuple(self.d.keys()),keys,"Keys from items() not equal to keys from keys()")
        self.assertEqual(tuple(self.d.values()),values,"Values from items() not equal to values from values()")
        self.assertEqual(self.d["Int"],1)
        self.assertEqual(self.d["Float"],1.0)
        self.assertEqual(self.d["Test"],self.d.metadata["Test"])
        self.assertEqual(self.d.metadata._typehints["Int"],"I32")
        self.assertEqual(len(self.d.dir()),6,"Failed meta data directory listing ({})".format(len(self.d.dir())))
        self.assertEqual(len(self.d3["Temperature"]),7,"Regular experssion metadata search failed")

    def test_dir(self):
        self.assertTrue(self.d.dir("S")==["Stoner.class"],"Dir method failed: dir was {}".format(self.d.dir()))
        bad_keys=set(['__metaclass__', 'iteritems', 'iterkeys', 'itervalues','__ge__', '__gt__', '__init_subclass__',
                      '__le__', '__lt__', '__reversed__', '__slots__',"_abc_negative_cache","_abc_registry",
                      "_abc_negative_cache_version","_abc_cache","_abc_impl"])
        self.attrs=set(dir(self.d))-bad_keys
        if len(self.attrs)!=230:
            expected={'_conv_string', '__str__', 'clear', 'scale', '__add__', 'popitem', '__mod_core__', 'priority', '_init_single', 'pop', 'reorder_columns', '_col_label', 'subclasses', '__sizeof__', 'rows', 'plot_matrix', '_PlotMixin__SurfPlotter', '_showfig', '__and__', '_repr_html_', '_pop_mask', 'filename', 'smooth', '__weakref__', 'dir', '_PlotMixin__mpl3DQuiver', 'spline', '__format__', 'plot_xy', 'labels', '_fix_kargs', 'span', '__getattr__', '_push_mask', 'normalise', 'mean', 'sort', '_set_mask', 'shape', 'x2', 'clone', 'save', 'setdefault', 'update', 'plot', 'colormap_xyz', 'plot_xyzuvw', '_patterns', 'section', 'count', 'extrapolate', 'fig', 'mime_type', 'find_col', 'SG_Filter', '__isub__', 'clip', '_fix_fig', '_AnalysisMixin__get_math_val', 'get_filename', 'griddata', 'axes', 'setas', '__sub__', '__floordiv__', '_baseclass', '_subplots', '__add_core__', 'max', 'rename', '__setstate__', 'rolling_window', '_labels', '__dict__', 'adjust_setas', 'column', '__mod__', 'xlim', '_repr_table_', '__abstractmethods__', '__call__', '_public_attrs_real', '__eq__', '_pyplot_proxy', 'stitch', 'lmfit', '_filename', '__lshift__', '_fix_titles', 'multiple', '__dir__', 'column_headers', 'header', '_public_attrs', 'swap_column', 'annotate_fit', '_init_many', 'min', '_load', '__sub_core__', 'dtype', '__doc__', '_col_args', 'filter', '__new__', '__len__', 'format', 'ylim', 'ax', '__hash__', '_PlotMixin__figure', 'polyfit', '__repr__', 'subtract', '__iand__', 'debug', 'diffsum', 'split', 'ylabel', '__iter__', '_vector_color', '__invert__', '__repr_core__', 'del_nan', 'y2', '__contains__', '__reduce__', 'plot_xyuv', 'plot_xyuvw', '__delattr__', 'curve_fit', '__module__', '_conv_float', '__getattribute__', 'keys', 'legend', 'quiver_plot', 'metadata', 'plot_xyz', '__regexp_meta__', 'data', 'figure', 'records', '_DataFile__parse_metadata', 'fignum', '__setattr__', 'insert_rows', 'add', 'make_bins', '_DataFile__setattr_col', '_repr_short_', '__getitem__', '_repr_limits', '_data', '_template', 'outlier_detection', 'template', '__imod__', 'image_plot', '__setitem__', '_get_curve_fit_data', '_DataFile__search_index', 'no_fmt', '__class__', '_AnalysisMixin__threshold', 'dims', 'threshold', 'basename', '_AnalysisMixin__lmfit_one', 'del_rows', 'patterns', 'del_column', '_metadata', '__deepcopy__', 'search', '_record_curve_fit_result', 'positional_fmt', '__getstate__', 'interpolate', 'dict_records', '_span_slice', 'columns', 'title', '__delitem__', '_repr_html_private', '_DataFile__file_dialog', '_getattr_col', 'mask', 'add_column', 'subplot', 'subplots', 'peaks', '_MutableMapping__marker', 'subplot2grid', 'get', '_DataFile__read_iterable', 'contour_xyz', 'inset', '__meta__', '_VectorFieldPlot', '__init__', '_masks', 'select', 'unique', 'xlabel', '_Plot', '__iadd__', 'values', 'multiply', '_raise_type_error', 'divide', 'odr', '__reduce_ex__', 'cmap', 'showfig', '__subclasshook__', 'items', '_interesting_cols', '_init_double', '__ne__', '_fix_cols', 'integrate', 'decompose', 'bin', 'closest', '__and_core__', 'T', 'load', 'apply'}
            print("Warning=====>",self.attrs-expected,expected-self.attrs)
        self.assertEqual(len(self.attrs),230,"DataFile.__dir__ failed.")

    def test_filter(self):
        self.d._push_mask()
        ix=np.argmax(self.d.x)
        self.d.filter(lambda r:r.x<=50)
        self.assertTrue(np.max(self.d.x)==50,"Failure of filter method to set mask")
        self.assertTrue(np.isnan(self.d.x[ix]),"Failed to mask maximum value")
        self.d._pop_mask()
        self.assertEqual(self.d2.select(Temp__not__gt=150).shape,(839,3),"Seect method failure.")
        self.assertEqual(self.d.select(lambda r:r.x<30),self.d.select(X__lt=30),"Select  method as callable failed.")
        self.assertEqual(self.d.select(__=lambda r:r.x<30),self.d.select(X__lt=30),"Select  method as callable failed.")

    def test_operators(self):
        #Test Column Indexer
        self.assertTrue(all(self.d//0==self.d.column(0)),"Failed the // operator with integer")
        self.assertTrue(all(self.d//"X-Data"==self.d.column(0)),"Failed the // operator with string")
        self.assertTrue(all((self.d//re.compile(r"^X\-"))[:,0]==self.d.column(0)),"Failed the // operator with regexp")
        t=[self.d%1,self.d%"Y-Data",(self.d%re.compile(r"Y\-"))]
        for ix,tst in enumerate(["integer","string","regexp"]):
            self.assertTrue(all(t[ix].data==np.atleast_2d(self.d.column(0)).T),"Failed % operator with {} index".format(tst))
        d=self.d&self.d.x
        self.assertTrue(d.shape[1]==3,"& operator failed.")
        self.assertTrue(len(d.setas)==3,"& messed up setas")
        self.assertTrue(len(d.column_headers)==3,"& messed up setas")
        d&=d.x
        self.assertTrue(d.shape[1]==4,"Inplace & operator failed")
        empty=Data()
        empty=empty+np.zeros(10)
        self.assertEqual(empty.shape,(1,10),"Adding to an empty array failed")
        d=self.d+np.array([0,0])
        self.assertTrue(len(d)==len(self.d)+1,"+ operator failed.")
        d+=np.array([0,0])
        self.assertTrue(len(d)==len(self.d)+2,"Inplace + operator failed.")
        d=d-(-1)
        self.assertTrue(len(d)==len(self.d)+1,"Delete with integer index failed. {} vs {}".format(len(d),len(self.d)+1))
        d-=-1
        self.assertTrue(len(d)==len(self.d),"Inplace delete with integer index failed. {} vs {}".format(len(d),len(self.d)))
        d-=slice(0,-1,2)
        self.assertTrue(len(d)==len(self.d)/2,"Inplace delete with slice index failed. {} vs {}".format(len(d),len(self.d)/2))
        d=d%1
        self.assertTrue(d.shape[1]==1,"Column division failed")
        self.assertTrue(len(d.setas)==d.shape[1],"Column removal messed up setas")
        self.assertTrue(len(d.column_headers)==d.shape[1],"Column removal messed up column headers")
        e=self.d.clone
        f=self.d2.clone
        g=e+f
        self.assertTrue(g.shape==(1776,2),"Add of 2 column and 3 column failed")
        g=f+e
        self.assertTrue(g.shape==(1776,3),"Add of 3 column and 2 column failed.")
        g=e&f
        h=f&e
        self.assertTrue(g.shape==h.shape,"Anding unequal column lengths faile!")
        e=~self.d
        self.assertTrue(e.setas[0]=="y","failed to invert setas columns")
        f.setas="xyz"

    def test_properties(self):
        self.little=Data()
        p=np.linspace(0,np.pi,91)
        q=np.linspace(0,2*np.pi,91)
        r=np.cos(p)
        x=r*np.sin(q)
        y=r*np.cos(q)
        self.little.data=np.column_stack((x,y,r))
        self.little.setas="xyz"
        q_ang=np.round(self.little.q/np.pi,decimals=2)
        p_ang=np.round(self.little.p/np.pi,decimals=2)
        self.assertTrue(np.max(q_ang)==1.0,"Data.q failure")
        self.assertTrue(np.max(p_ang)==0.5,"Data.p failure")
        self.assertTrue(np.min(p_ang)==-0.5,"Data.p failure")
        self.little.mask=mask_func
        self.little.mask[:,2]=True
        self.assertEqual(len(repr(self.little).split("\n")),len(self.little)+5,"Non shorterned representation failed.")

    def test_methods(self):
        d=self.d.clone
        d&=np.where(d.x<50,1.0,0.0)
        d.rename(2,"Z-Data")
        d.setas="xyz"
        self.assertTrue(all(d.unique(2)==np.array([0,1])),"Unique values failed: {}".format(d.unique(2)))
        d=self.d.clone
        d.insert_rows(10,np.zeros((2,2)))
        self.assertEqual(len(d),102,"Failed to inert extra rows")
        self.assertTrue(d[9,0]==10 and d[10,0]==0 and d[12,0]==11, "Failed to insert rows properly.")
        d=self.d.clone
        d.add_column(np.ones(len(d)),replace=False,header="added")
        self.assertTrue(d.shape[1]==self.d.shape[1]+1,"Adding a column with replace=False did not add a column.")
        self.assertTrue(np.all(d.data[:,-1]==np.ones(len(d))),"Didn't add the new column to the end of the data.")
        self.assertTrue(len(d.column_headers)==len(self.d.column_headers)+1,"Column headers isn't bigger by one")
        self.assertTrue(d.column_headers==self.d.column_headers+["added",],"Column header not added correctly")
        d=self.d.clone
        d.add_column(self.d.x)
        self.assertTrue(np.all(d.x==d[:,2]),"Adding a column as a DataArray with column headers didn't work")
        self.assertTrue(d.x.column_headers[0]==d.column_headers[2],"Adding a column as a DataArray with column headers didn't work")
        e=d.clone
        d.swap_column([(0,1),(0,2)])
        self.assertTrue(d.column_headers==[e.column_headers[x] for x in [2,0,1]],"Swap column test failed: {}".format(d.column_headers))
        e=self.d(setas="yx")
        self.assertTrue(e.shape==self.d.shape and e.setas[0]=="y","Failed on a DataFile.__call__ test")
        spl=len(repr(self.d).split("\n"))
        self.assertEqual(spl,105,"Failed to do repr function got {} lines".format(spl))
        e=self.d.clone
        e=e.add_column(e.x,header=e.column_headers[0])
        e.del_column(duplicates=True)
        self.assertTrue(e.shape==(100,2),"Deleting duplicate columns failed")
        e=self.d2.clone
        e.reorder_columns([2,0,1])
        self.assertTrue(e.column_headers==[self.d2.column_headers[x] for x in [2,0,1]],"Failed to reorder columns: {}".format(e.column_headers))
        d=self.d.clone
        d.del_rows(0,10.0)
        self.assertEqual(d.shape,(99,2),"Del Rows with value and column failed - actual shape {}".format(d.shape))
        d=self.d.clone
        d.del_rows(0,(10.0,20.0))
        self.assertEqual(d.shape,(89,2),"Del Rows with tuple and column failed - actual shape {}".format(d.shape))
        d=self.d.clone
        d.mask[::2,0]=True
        d.del_rows()
        self.assertEqual(d.shape,(50,2),"Del Rows with mask set - actual shape {}".format(d.shape))
        d=self.d.clone
        d[::2,1]=nan
        d.del_nan(1)
        self.assertEqual(d.shape,(50,2),"del_nan with explicit column set failed shape was {}".format(d.shape))
        d=self.d.clone
        d[::2,1]=nan
        d.del_nan(0)
        self.assertEqual(d.shape,(100,2),"del_nan with explicit column set and not nans failed shape was {}".format(d.shape))
        d=self.d.clone
        d[::2,1]=nan
        d.setas=".y"
        d.del_nan()
        self.assertEqual(d.shape,(50,2),"del_nan with columns from setas failed shape was {}".format(d.shape))
        d=self.d.clone
        d2=self.d.clone
        d2.data=d2.data[::-1,:]
        self.assertEqual(d.sort(reverse=True),d2,"Sorting revserse not the same as manually reversed data.")

    def test_metadata_save(self):
        local = path.dirname(__file__)
        t = np.arange(12).reshape(3,4) #set up a test data file with mixed metadata
        t = Data(t)
        t.column_headers = ["1","2","3","4"]
        metitems = [True,1,0.2,{"a":1, "b":"abc"},(1,2),np.arange(3),[1,2,3], "abc", #all types accepted
                    r"\\abc\cde", 1e-20, #extra tests
                    [1,(1,2),"abc"], #list with different types
                    [[[1]]] #nested list
                    ]
        metnames = ["t"+str(i) for i in range(len(metitems))]
        for k,v in zip(metnames,metitems):
            t[k] = v
        t.save(path.join(local, "mixedmetatest.dat"))
        tl = Data(path.join(local, "mixedmetatest.txt")) #will change extension to txt if not txt or tdi, is this what we want?
        t2 = self.d4.clone  #check that python tdi save is the same as labview tdi save
        t2.save(path.join(local, "mixedmetatest2.txt"))
        t2l = Data(path.join(local, "mixedmetatest2.txt"))
        for orig, load in [(t,tl), (t2, t2l)]:
            self.assertTrue(np.allclose(orig.data, load.data))
            self.assertTrue(orig.column_headers==load.column_headers)
            self.assertTrue(all([i in load.metadata.keys() for i in orig.metadata.keys()]))
            for k in orig.metadata.keys():
                if isinstance(orig[k], np.ndarray):
                    self.assertTrue(np.allclose(load[k],orig[k]))
                elif isinstance(orig[k], float) and np.isnan(orig[k]):
                    self.assertTrue(np.isnan(load[k]))
                else:
                    self.assertTrue(load[k] == orig[k], "Not equal for metadata: {}".format(load[k]))
        os.remove(path.join(local, "mixedmetatest.txt")) #clear up
        os.remove(path.join(local, "mixedmetatest2.txt"))

if __name__=="__main__": # Run some tests manually to allow debugging
    test=Datatest("test_operators")
    test.setUp()
    #test.test_properties()
    #test.test_methods()
    #test.test_filter()
#    test.test_deltions()
    #test.test_dir()
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec  8 15:20:15 2018

@author: phygbu
"""

import unittest
import sys
import  os.path as path

pth=path.dirname(__file__)
pth=path.realpath(path.join(pth,"../../"))
sys.path.insert(0,pth)
from Stoner.core.base import typeHintedDict

class typeHintedDictTest(unittest.TestCase):
    """Test typeHintedDict class"""

    def test_ops(self):
        d = typeHintedDict([('el1',1),('el2',2),('el3',3),('other',4)])
        d.filter('el')
        self.assertTrue(len(d)==3)
        d.filter(lambda x: x.endswith('3'))
        self.assertTrue(len(d)==1)
        self.assertTrue(d['el3']==3)
        d["munge"]=None
        self.assertTrue(d.types["munge"]=="String","Setting type for None value failed.")
        d["munge"]=1
        self.assertTrue(d["munge{String}"]=="1","Munging return type in getitem failed.")
        self.assertEqual(repr(d),
        """'el3':I32:3
'munge':I32:1""","Repr failed \n{}".format(d))


if __name__=="__main__": # Run some tests manually to allow debugging
    test=typeHintedDictTest("test_ops")
    test.setUp()
    #test.test_properties()
    #test.test_methods()
    #test.test_filter()
#    test.test_deltions()
    #test.test_dir()
    unittest.main()



import unittest
import sys
import os.path as path
import os
from Stoner.compat import listdir_recursive
from importlib import import_module
import matplotlib.pyplot as plt
from traceback import format_exc

pth=path.dirname(__file__)
pth=path.realpath(path.join(pth,"../../"))
sys.path.insert(0,pth)

skip_scipts=["plot-folder-test"]

class DocSamples_test(unittest.TestCase):

    """Path to sample Data File"""
    datadir=path.join(pth,"doc","samples")

    def setUp(self):
        self.scripts=[path.relpath(x,self.datadir).replace(path.sep,".") for x in listdir_recursive(self.datadir,"*.py") if not x.endswith("__init__.py")]
        sys.path.insert(0,self.datadir)

    def test_scripts(self):
        """Import each of the sample scripts in turn and see if they ran without error"""
        failures=[]
        for ix,filename in enumerate(self.scripts):
            os.chdir(self.datadir)
            script=filename[:-3]
            if script in skip_scipts:
                print("Skippoing {}".format(filename))
                continue
            print("Trying script {}: {}".format(ix,filename))
            try:
                os.chdir(self.datadir)
                code=import_module(script)
                fignum=len(plt.get_fignums())
                self.assertGreaterEqual(fignum,1,"{} Did not produce any figures !".format(script))
                print("Done")
                fig=plt.gcf()
                plt.close(fig)
                plt.close("all")
            except Exception:
                v=format_exc()
                print("Failed with\n{}".format(v))
                failures.append("Script file {} failed with {}".format(filename,v))
        self.assertTrue(len(failures)==0,"\n".join(failures))

if __name__=="__main__": # Run some tests manually to allow debugging
    test=DocSamples_test("test_scripts")
    test.setUp()
    test.test_scripts()





"""Stoner.Folders : Classes for working collections of :class:`.Data` files.

The core classes provides a means to access them as an ordered collection or as a mapping.
"""
__all__ = ["DataFolder", "PlotFolder"]

from .folders.core import baseFolder
from .folders.mixins import DiskBasedFolder, DataMethodsMixin, PlotMethodsMixin


class DataFolder(DataMethodsMixin, DiskBasedFolder, baseFolder):

    """Provide an interface to manipulating lots of data files stored within a directory structure on disc.

    By default, the members of the DataFolder are isntances of :class:`Stoner.Data`. The DataFolder emplys a lazy open strategy, so that
    files are only read in from disc when actually needed.

    .. inheritance-diagram:: DataFolder

    """

    def __init__(self, *args, **kargs):
        from Stoner import Data

        self.type = kargs.pop("type", Data)
        super(DataFolder, self).__init__(*args, **kargs)


class PlotFolder(PlotMethodsMixin, DataFolder):

    """A :py:class:`Stoner.folders.baseFolder` that knows how to ploth its underlying data files."""



# -*- coding: utf-8 -*-
"""
test_Core.py
Created on Tue Jan 07 22:05:55 2014

@author: phygbu
"""

import unittest
import sys
import os, os.path as path
import numpy as np
import re
from numpy import any,all,sqrt,nan
from matplotlib.figure import Figure
import matplotlib.pyplot as plt

pth=path.dirname(__file__)
pth=path.realpath(path.join(pth,"../../../"))
sys.path.insert(0,pth)
from Stoner import Data,__home__,Options
from Stoner.Core import typeHintedDict

from Stoner.plot.formats import DefaultPlotStyle

class Plottest(unittest.TestCase):

    """Path to sample Data File"""
    datadir=path.join(pth,"sample-data")

    def setUp(self):
        self.d=Data(path.join(__home__,"..","sample-data","New-XRay-Data.dql"))

    def test_set_no_figs(self):
        self.assertTrue(Options.no_figs,"Default setting for no_figs option is incorrect.")
        Options.no_figs=True
        e=self.d.clone
        ret=e.plot()
        self.assertTrue(ret is None,"Output of Data.plot() was not None when no_figs is True  and showfig is not set({})".format(type(ret)))
        Options.no_figs=False
        e.showfig=False
        ret=e.plot()
        self.assertTrue(isinstance(ret,Data),"Return value of Data.plot() was not self when Data.showfig=False ({})".format(type(ret)))
        e.showfig=True
        ret=e.plot()
        self.assertTrue(isinstance(ret,Figure),"Return value of Data.plot() was not Figure when Data.showfig=False({})".format(type(ret)))
        e.showfig=None
        ret=e.plot()
        self.assertTrue(ret is None,"Output of Data.plot() was not None when Data.showfig is None ({})".format(type(ret)))
        Options.no_figs=True
        self.assertTrue(Options.no_figs,"set_option no_figs failed.")
        self.d=Data(path.join(__home__,"..","sample-data","New-XRay-Data.dql"))
        self.d.showfig=False
        ret=self.d.plot()
        self.assertTrue(ret is None,"Output of Data.plot() was not None when no_figs is True ({})".format(type(ret)))
        Options.no_figs=True
        plt.close("all")

    def test_template_settings(self):
        template=DefaultPlotStyle(font__weight="bold")
        self.assertEqual(template["font.weight"],"bold","Setting ytemplate parameter in init failed.")
        template(font__weight="normal")
        self.assertEqual(template["font.weight"],"normal","Setting ytemplate parameter in call failed.")
        template["font.weight"]="bold"
        self.assertEqual(template["font.weight"],"bold","Setting ytemplate parameter in setitem failed.")
        del template["font.weight"]
        self.assertEqual(template["font.weight"],"normal","Resettting template parameter failed.")
        keys=sorted([x for x in template])
        self.assertEqual(sorted(template.keys()),keys,"template.keys() and template.iter() disagree.")
        attrs=[x for x in dir(template) if template._allowed_attr(x)]
        length=len(dict(plt.rcParams))+len(attrs)
        self.assertEqual(len(template),length,"templa length wrong.")






if __name__=="__main__": # Run some tests manually to allow debugging
    test=Plottest("test_set_no_figs")
    test.setUp()
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Implements DataFile like classes for various large scale facilities
"""
__all__ = ["BNLFile", "MDAASCIIFile", "OpenGDAFile", "RasorFile", "SNSFile"]
# Standard Library imports
import io
import linecache
import re

import numpy as np

import Stoner.Core as Core
from Stoner.compat import python_v3, str2bytes


class BNLFile(Core.DataFile):

    """
    Creates BNLFile a subclass of Core.DataFile that caters for files in the SPEC format given by BNL (specifically u4b beamline but hopefully generalisable).

    Author Rowan 12/2011

    The file from BNL must be split into seperate scan files before Stoner can use
    them, a separate python script has been written for this and should be found
    in data/Python/PythonCode/scripts.
    """

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 64
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.txt"]  # Recognised filename patterns

    def __init__(self, *params):
        """Constructor modification.

        Do a normal initiation using the parent class 'self' followed by adding an extra attribute line_numbers,
        line_numbers is a list of important line numbers in the file.
        I've left it open for someone to add options for more args if they wish.
        """
        super(BNLFile, self).__init__(*params)
        self.line_numbers = []

    def __find_lines(self):
        """Returns an array of ints [header_line,data_line,scan_line,date_line,motor_line]."""
        with io.open(self.filename, "r", errors="ignore", encoding="utf-8") as fp:
            self.line_numbers = [0, 0, 0, 0, 0]
            counter = 0
            for line in fp:
                counter += 1
                if counter == 1 and line[0] != "#":
                    raise Core.StonerLoadError("Not a BNL File ?")
                if len(line) < 2:
                    continue  # if there's nothing written on the line go to the next
                elif line[0:2] == "#L":
                    self.line_numbers[0] = counter
                elif line[0:2] == "#S":
                    self.line_numbers[2] = counter
                elif line[0:2] == "#D":
                    self.line_numbers[3] = counter
                elif line[0:2] == "#P":
                    self.line_numbers[4] = counter
                elif line[0] in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]:
                    self.line_numbers[1] = counter
                    break

    def __get_metadata(self):
        """Load metadta from file.

        Metadata found is scan number 'Snumber', scan type and parameters 'Stype',
        scan date/time 'Sdatetime' and z motor position 'Smotor'.
        """
        scanLine = linecache.getline(self.filename, self.line_numbers[2])
        dateLine = linecache.getline(self.filename, self.line_numbers[3])
        motorLine = linecache.getline(self.filename, self.line_numbers[4])
        self.__setitem__("Snumber", scanLine.split()[1])
        tmp = "".join(scanLine.split()[2:])
        self.__setitem__("Stype", "".join(tmp.split(",")))  # get rid of commas
        self.__setitem__("Sdatetime", dateLine[3:-1])  # don't want \n at end of line so use -1
        self.__setitem__("Smotor", motorLine.split()[3])

    def __parse_BNL_data(self):
        """Internal function for parsing BNL data.

         The meta data is labelled by #L type tags
        so easy to find but #L must be excluded from the result.
        """
        self.__find_lines()
        # creates a list, line_numbers, formatted [header_line,data_line,scan_line,date_line,motor_line]
        header_string = linecache.getline(self.filename, self.line_numbers[0])
        header_string = re.sub(r'["\n]', "", header_string)  # get rid of new line character
        header_string = re.sub(r"#L", "", header_string)  # get rid of line indicator character
        column_headers = map(lambda x: x.strip(), header_string.split())
        self.__get_metadata()
        try:
            self.data = np.genfromtxt(self.filename, skip_header=self.line_numbers[1] - 1)
        except IOError:
            self.data = np.array([0])
            print("Did not import any data for {}".format(self.filename))
        self.column_headers = column_headers

    def _load(self, filename, *args, **kargs):  # fileType omitted, implicit in class call
        """BNLFile.load(filename)

        Args:
            filename (string or bool): File to load. If None then the existing filename is used,
                if False, then a file dialog will be used.

        Returns:
            A copy of the itself after loading the data.

        Notes:
            Overwrites load method in Core.DataFile class, no header positions and data
            positions are needed because of the hash title structure used in BNL files.

            Normally its good to use _parse_plain_data method from Core.DataFile class
            to load data but unfortunately Brookhaven data isn't very plain so there's
            a new method below.
        """
        self.filename = filename
        self.__parse_BNL_data()  # call an internal function rather than put it in load function
        linecache.clearcache()
        return self


class MDAASCIIFile(Core.DataFile):

    """Reads files generated from the APS."""

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 16
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.txt"]  # Recognised filename patterns

    def _load(self, filename=None, *args, **kargs):
        """Load function. File format has space delimited columns from row 3 onwards."""
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename
        i = [0, 0, 0, 0]
        with io.open(self.filename, "r", errors="ignore", encoding="utf-8") as data:  # Slightly ugly text handling
            for i[0], line in enumerate(data):
                if (
                    i[0] == 0 and line.strip() != "## mda2ascii 1.2 generated output"
                ):  # bug out oif we don't like the header
                    raise Core.StonerLoadError("Not a file mda2ascii")
                line.strip()
                if "=" in line:
                    parts = line[2:].split("=")
                    self[parts[0].strip()] = self.metadata.string_to_type("".join(parts[1:]).strip())
                elif line.startswith("#  Extra PV:"):
                    # Onto the next metadata bit
                    break
            pvpat = re.compile(r"^#\s+Extra\s+PV\s\d+\:(.*)")
            for i[1], line in enumerate(data):
                if line.strip() == "":
                    continue
                elif line.startswith("# Extra PV"):
                    res = pvpat.match(line)
                    bits = [b.strip().strip(r'"') for b in res.group(1).split(",")]
                    if bits[1] == "":
                        key = bits[0]
                    else:
                        key = bits[1]
                    if len(bits) > 3:
                        key = key + " ({})".format(bits[3])
                    self[key] = self.metadata.string_to_type(bits[2])
                else:
                    break  # End of Extra PV stuff
            else:
                raise Core.StonerLoadError("Overran Extra PV Block")
            for i[2], line in enumerate(data):
                line.strip()
                if line.strip() == "":
                    continue
                elif line.startswith("# Column Descriptions:"):
                    break  # Start of column headers now
                elif "=" in line:
                    parts = line[2:].split("=")
                    self[parts[0].strip()] = self.metadata.string_to_type("".join(parts[1:]).strip())
            else:
                raise Core.StonerLoadError("Overran end of scan header before column descriptions")
            colpat = re.compile(r"#\s+\d+\s+\[([^\]]*)\](.*)")
            column_headers = []
            for i[3], line in enumerate(data):
                res = colpat.match(line)
                line.strip()
                if line.strip() == "":
                    continue
                elif line.startswith("# 1-D Scan Values"):
                    break  # Start of data
                elif res is not None:
                    if "," in res.group(2):
                        bits = [b.strip() for b in res.group(2).split(",")]
                        if bits[-2] == "":
                            colname = bits[0]
                        else:
                            colname = bits[-2]
                        if bits[-1] != "":
                            colname += " ({})".format(bits[-1])
                        if colname in column_headers:
                            colname = "{}:{}".format(bits[0], colname)
                    else:
                        colname = res.group(1).strip()
                    column_headers.append(colname)
            else:
                raise Core.StonerLoadError("Overand the end of file without reading data")
        self.data = np.genfromtxt(self.filename, skip_header=sum(i))  # so that's ok then !
        self.column_headers = column_headers
        return self


class OpenGDAFile(Core.DataFile):

    """Extends Core.DataFile to load files from RASOR"""

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 16  # Makes a positive ID of it's file type so give priority
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.dat"]  # Recognised filename patterns

    def _load(self, filename=None, *args, **kargs):
        """Load an OpenGDA file.

        Args:
            filename (string or bool): File to load. If None then the existing filename is used,
                if False, then a file dialog will be used.

        Returns:
            A copy of the itself after loading the data.
        """
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename
        i = 0
        with io.open(self.filename, "r", errors="ignore", encoding="utf-8") as f:
            for i, line in enumerate(f):
                line = line.strip()
                if i == 0 and line != "&SRS":
                    raise Core.StonerLoadError("Not a GDA File from Rasor ?" + str(line))
                if "&END" in line:
                    break
                parts = line.split("=")
                if len(parts) != 2:
                    continue
                key = parts[0]
                value = parts[1].strip()
                self.metadata[key] = self.metadata.string_to_type(value)
            if python_v3:
                column_headers = f.readline().strip().split("\t")
            else:
                column_headers = f.next().strip().split("\t")
            self.data = np.genfromtxt([str2bytes(l) for l in f], dtype="float", invalid_raise=False)
        self.column_headers = column_headers
        return self


class RasorFile(OpenGDAFile):

    """Just an alias for OpenGDAFile"""

    pass


class SNSFile(Core.DataFile):

    """Reads the ASCII exported Poalrised Neutron Rfeflectivity reduced files from BL-4A line at the Spallation Neutron Source at Oak Ridge National Lab.

    File has a large header marked up with # prefixes which include several section is []
    Each section seems to have a slightly different format
    """

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 16
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.dat"]  # Recognised filename patterns

    def _load(self, filename=None, *args, **kargs):
        """Load function. File format has space delimited columns from row 3 onwards."""
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename

        with io.open(self.filename, "r", errors="ignore", encoding="utf-8") as data:  # Slightly ugly text handling
            line = data.readline()
            if line.strip() != "# datafile created by QuickNXS 0.9.39":  # bug out oif we don't like the header
                raise Core.StonerLoadError("Not a file from the SNS BL4A line")
            for line in data:
                if line.startswith("# "):  # We're in the header
                    line = line[2:].strip()  # strip the header and whitespace

                if line.startswith("["):  # Look for a section header
                    section = line.strip().strip("[]")
                    if section == "Data":  # The Data section has one line of colum headers and then data
                        header = next(data)[2:].split("\t")
                        column_headers = [h.strip().decode("ascii", "ignore") for h in header]
                        self.data = np.genfromtxt(data)  # we end by reading the raw data
                    elif section == "Global Options":  # This section can go into metadata
                        for line in data:
                            line = line[2:].strip()
                            if line.strip() == "":
                                break
                            else:
                                self[line[2:10].strip()] = line[11:].strip()
                    elif (
                        section == "Direct Beam Runs" or section == "Data Runs"
                    ):  # These are constructed into lists ofg dictionaries for each file
                        sec = list()
                        header = next(data)
                        header = header[2:].strip()
                        keys = [s.strip() for s in header.split("  ") if s.strip()]
                        for line in data:
                            line = line[2:].strip()
                            if line == "":
                                break
                            else:
                                values = [s.strip() for s in line.split("  ") if s.strip()]
                                sec.append(dict(zip(keys, values)))
                        self[section] = sec
                else:  # We must still be in the opening un-labelled section of meta data
                    if ":" in line:
                        i = line.index(":")
                        key = line[:i].strip()
                        value = line[i + 1 :].strip()
                        self[key.strip()] = value.strip()
        self.column_headers = column_headers
        return self



# -*- coding: utf-8 -*-
"""Implements a baseFolder type structure for working with collections of images."""
__all__ = ["_generator", "ImageFolderMixin", "ImageFolder"]
from .core import ImageArray
from Stoner import Data
from Stoner.Folders import DiskBasedFolder, baseFolder
from Stoner.compat import string_types
from Stoner.tools import isiterable, islike_list
from Stoner.Image import ImageFile, ImageArray

from skimage.viewer import CollectionViewer
import numpy as np


def _load_ImageArray(f, **kargs):
    """Simple meothd to load an image array."""
    kargs.pop("img_num", None)
    return ImageArray(f, **kargs)


class _generator(object):

    """A helper class to iterator over ImageFolder yet remember it's own length."""

    def __init__(self, fldr):
        self.fldr = fldr
        self.len = len(fldr)

    def __len__(self):
        return self.len

    def __iter__(self):
        self.ix = 0
        return self

    def __next__(self):
        if self.ix < len(self):
            ret = self[self.ix]
            self.ix += 1
            return ret
        else:
            raise StopIteration("Finished iterating Folder.")

    def __getitem__(self, index):
        ret = self.fldr[index]
        if hasattr(ret, "image"):
            ret = ret.image
        return ret

    def next(self):
        return self.__next__()


class ImageFolderMixin(object):

    """Mixin to provide a folder object for images.

    ImageFolderMixin is designed to behave pretty much like DataFolder but with
    functions and loaders appropriate for image based files.

        Attributes:
        type (:py:class:`Stoner.Image.core.ImageArray`) the type ob object to sotre in the folder (defaults to :py:class:`Stoner.Cire.Data`)

        extra_args (dict): Extra arguments to use when instantiatoing the contents of the folder from a file on disk.

        pattern (str or regexp): A filename globbing pattern that matches the contents of the folder. If a regular expression is provided then
            any named groups are used to construct additional metadata entryies from the filename. Default is *.* to match all files with an extension.

        read_means (bool): IF true, additional metatdata keys are added that return the mean value of each column of the data. This can hep in
            grouping files where one column of data contains a constant value for the experimental state. Default is False

        recursive (bool): Specifies whether to search recurisvely in a whole directory tree. Default is True.

        flatten (bool): Specify where to present subdirectories as spearate groups in the folder (False) or as a single group (True). Default is False.
            The :py:meth:`DiskBasedFolder.flatten` method has the equivalent effect and :py:meth:`DiskBasedFolder.unflatten` reverses it.

        directory (str): The root directory on disc for the folder - by default this is the current working directory.

        multifile (boo): Whether to select individual files manually that are not (necessarily) in  a common directory structure.

        readlist (bool): Whether to read the directory immediately on creation. Default is True
    """

    _defaults = {"type": ImageArray, "pattern": ["*.png", "*.tiff", "*.jpeg", "*.jpg", "*.tif"]}
    _no_defaults = ["flat"]

    def __init__(self, *args, **kargs):
        """nitialise the ImageFolder.

        Mostly a pass through to the :py:class:`Stoner.Folders.baseFolder` class.
        """
        super(ImageFolderMixin, self).__init__(*args, **kargs)

    @property
    def size(self):
        """Return the size of an individual image or False if not all images are the same size."""
        shape = self.images[0].shape
        for i in self.images:
            if i.shape != shape:
                return False
        return shape

    @property
    def images(self):
        """A generator that iterates over just the images in the Folder."""
        return _generator(self)

    def _getattr_proxy(self, item):
        """Override baseFolder proxy call to access a method of the ImageFile

        Args:
            item (string): Name of method of metadataObject class to be called

        Returns:
            Either a modifed copy of this objectFolder or a list of return values
            from evaluating the method for each file in the Folder.
        """
        meth = getattr(self.instance, item, None)

        def _wrapper_(*args, **kargs):
            """Wraps a call to the metadataObject type for magic method calling.

            Keyword Arguments:
                _return (index types or None): specify to store the return value in the individual object's metadata

            Note:
                This relies on being defined inside the enclosure of the objectFolder method
                so we have access to self and item
            """
            retvals = []
            _return = kargs.pop("_return", None)
            for ix, f in enumerate(self):
                meth = getattr(f, item, None)
                ret = meth(*args, **kargs)  # overwriting array is handled by ImageFile proxy function
                retvals.append(ret)
                if item == "crop":
                    self[ix] = ret
                if _return is not None:
                    if isinstance(_return, bool) and _return:
                        _return = meth.__name__
                    self[ix][_return] = ret
            return retvals

        # Ok that's the wrapper function, now return  it for the user to mess around with.
        _wrapper_.__doc__ = meth.__doc__
        _wrapper_.__name__ = meth.__name__
        return _wrapper_

    def apply_all(self, *args, **kargs):
        """apply function to all images in the stack

        Args:
            func(string or callable):
                if string it must be a function reachable by ImageArray
            quiet(bool):
                if False print '.' for every iteration

        Note:
            Further args, kargs are passed through to the function
        """
        args = list(args)
        func = args.pop(0)
        quiet = kargs.pop("quiet", True)
        if isinstance(func, string_types):
            for i, im in enumerate(self):
                f = getattr(im, func)
                self[i] = f(*args, **kargs)
                if not quiet:
                    print(".")
        elif hasattr(func, "__call__"):
            for i, im in enumerate(self):
                self[i] = func(im, *args, **kargs)
            if not quiet:
                print(".")

    def average(self, weights=None, _box=None):
        """Get an array of average pixel values for the stack.

        Pass through to numpy average
        Returns:
            average(ImageArray):
                average values
        """
        if not self.size:
            raise RuntimeError("Cannot average Imagefolder if images have different sizes")
        stack = np.stack(self.images, axis=0)
        average = np.average(stack, axis=0, weights=weights)
        ret = average.view(ImageArray)
        ret.metadata = self.metadata.common_metadata
        return ImageFile(ret[ret._box(_box)])

    def loadgroup(self):
        """Load all files from this group into memory"""
        for _ in self:
            pass

    def as_stack(self):
        """Return a ImageStack of the images in the current group."""
        from Stoner.Image import ImageStack

        k = ImageStack(self)
        return k

    def mean(self, _box=None):
        """Calculate the mean value of all the images in the stack.

        Actually a synonym for self.average with not weights
        """
        return self.average(_box=_box)

    def stddev(self, weights=None):
        """Calculate weighted standard deviation for stack

        This is a biased standard deviation, may not be appropriate for small sample sizes
        """
        weights = np.ones(len(self)) if weights is None else weights
        avs = self.average(weights=weights)
        sumsqdev = np.zeros_like(avs)
        for ix, img in enumerate(self.images):
            sumsqdev += weights[ix] * (img - avs) ** 2
        sumsqdev = np.sqrt(sumsqdev) / np.sum(weights, axis=0)
        return sumsqdev.view(ImageArray)

    def stderr(self, weights=None):
        """Standard error in the stack average"""
        serr = self.stddev(weights=weights) / np.sqrt(len(self))
        return serr

    def view(self):
        """Create a matplotlib animated view of the contents.

        """
        cv = CollectionViewer(self.images)
        cv.show()
        return cv


class ImageFolder(ImageFolderMixin, DiskBasedFolder, baseFolder):

    """Folder object for images.

    ImageFolder is designed to behave pretty much like DataFolder but with
    functions and loaders appropriate for image based files.

        Attributes:
        type (:py:class:`Stoner.Image.core.ImageArray`) the type ob object to sotre in the folder (defaults to :py:class:`Stoner.Cire.Data`)

        extra_args (dict): Extra arguments to use when instantiatoing the contents of the folder from a file on disk.

        pattern (str or regexp): A filename globbing pattern that matches the contents of the folder. If a regular expression is provided then
            any named groups are used to construct additional metadata entryies from the filename. Default is *.* to match all files with an extension.

        read_means (bool): IF true, additional metatdata keys are added that return the mean value of each column of the data. This can hep in
            grouping files where one column of data contains a constant value for the experimental state. Default is False

        recursive (bool): Specifies whether to search recurisvely in a whole directory tree. Default is True.

        flatten (bool): Specify where to present subdirectories as spearate groups in the folder (False) or as a single group (True). Default is False.
            The :py:meth:`DiskBasedFolder.flatten` method has the equivalent effect and :py:meth:`DiskBasedFolder.unflatten` reverses it.

        directory (str): The root directory on disc for the folder - by default this is the current working directory.

        multifile (boo): Whether to select individual files manually that are not (necessarily) in  a common directory structure.

        readlist (bool): Whether to read the directory immediately on creation. Default is True
    """

    pass



# -*- coding: utf-8 -*-
"""
test_Folders.py

Created on Mon Jul 18 14:13:39 2016

@author: phygbu
"""


import unittest
import sys
import os.path as path
import os
import numpy as np
import re
import fnmatch
from numpy import ceil
from Stoner.compat import *
import Stoner.Folders as SF

from Stoner import Data,set_option
import Stoner.HDF5, Stoner.Zip
from Stoner.Util import hysteresis_correct

import matplotlib.pyplot as plt

import tempfile

pth=path.dirname(__file__)
pth=path.realpath(path.join(pth,"../../../"))
sys.path.insert(0,pth)

class Folders_test(unittest.TestCase):

    """Path to sample Data File"""
    datadir=path.join(pth,"sample-data")

    def setUp(self):
        pass

    def test_Folders(self):
        self.fldr=SF.DataFolder(self.datadir,debug=False,recursive=False)
        fldr=self.fldr
        fl=len(fldr)
        datfiles=fnmatch.filter(os.listdir(self.datadir),"*.dat")
        length = len([i for i in os.listdir(self.datadir) if path.isfile(os.path.join(self.datadir,i))])-1 # don't coiunt TDMS index
        self.assertEqual(length,fl,"Failed to initialise DataFolder from sample data")
        self.assertEqual(fldr.index(path.basename(fldr[-1].filename)),fl-1,"Failed to index back on filename")
        self.assertEqual(fldr.count(path.basename(fldr[-1].filename)),1,"Failed to count filename with string")
        self.assertEqual(fldr.count("*.dat"),len(datfiles),"Count with a glob pattern failed")
        self.assertEqual(len(fldr[::2]),ceil(len(fldr)/2.0),"Failed to get the correct number of elements in a folder slice")

    def test_loader_opts(self):
        self.fldr7=SF.DataFolder(path.join(self.datadir,"NLIV"),pattern=re.compile(r".*at (?P<field>[0-9\-\.]*)\.txt"),read_means=True)
        x=self.fldr7.metadata.slice(["field","Voltage","Current"],output="Data")
        self.assertEqual(x.span("field"),(-0.05,0.04),"Extract from name pattern and slice into metadata failed.")
        self.assertTrue(all(x//"Current"<0) and all(x//"Current">-1E-20),"Extract means failed.")
        self.assertEqual(list(self.fldr7.not_loaded),[],"Not loaded attribute failed.")
        self.fldr7.unload(0)
        self.assertEqual(len(list(self.fldr7.not_loaded)),1,"Unload by index failed.")
        self.fldr7.unload()
        self.assertEqual(len(list(self.fldr7.not_loaded)),len(self.fldr7),"Unload all failed.")
        def add_col(d):
            d.add_column(np.ones(len(d))*d["field"],header="field")
        self.fldr7.each(add_col)
        self.fldr7.concatenate()
        self.assertEqual(self.fldr7[0].shape,(909, 4),"Concatenate failed.")


    def test_discard_earlier(self):
        fldr2=SF.DataFolder(path.join(pth,"tests/Stoner/folder_data"),pattern="*.dat",discard_earlier=True)
        fldr3=SF.DataFolder(path.join(pth,"tests/Stoner/folder_data"),pattern="*.dat")
        self.assertEqual(len(fldr2),1,"Folder created with disacrd_earlier has wrong length ({})".format(len(fldr2)))
        self.assertEqual(len(fldr3),5,"Folder created without disacrd_earlier has wrong length ({})".format(len(fldr3)))
        fldr3.keep_latest()
        self.assertEqual(list(fldr2.ls),list(fldr3.ls),"Folder.keep_latest didn't do the same as discard_earliest in constructor.")

    def test_Operators(self):
        fldr=SF.DataFolder(self.datadir,debug=False,recursive=False)
        fl=len(fldr)
        d=Data(np.ones((100,5)))
        fldr+=d
        self.assertEqual(fl+1,len(fldr),"Failed += operator on DataFolder")
        fldr2=fldr+fldr
        self.assertEqual((fl+1)*2,len(fldr2),"Failed + operator with DataFolder on DataFolder")
        fldr-="Untitled"
        self.assertEqual(len(fldr),fl,"Failed to remove Untitled-0 from DataFolder by name.")
        fldr-="New-XRay-Data.dql"
        self.assertEqual(fl-1,len(fldr),"Failed to remove NEw Xray data by name.")
        fldr+="New-XRay-Data.dql"
        self.assertEqual(len(fldr),fl,"Failed += oeprator with string on DataFolder")
        fldr/="Loaded as"
        self.assertEqual(len(fldr["QDFile"]),4,"Failoed to group folder by Loaded As metadata with /= opeator.")

    def test_Properties(self):
        fldr=SF.DataFolder(self.datadir,debug=False,recursive=False)
        fldr/="Loaded as"
        fldr["QDFile"].group("Byapp")
        self.assertEqual(fldr.mindepth,1,"mindepth attribute of folder failed.")
        self.assertEqual(fldr.depth,2,"depth attribute failed.")
        fldr=SF.DataFolder(self.datadir,debug=False,recursive=False)
        fldr+=Data()
        self.assertEqual(len(list(fldr.loaded)),1,"loaded attribute failed {}".format(len(list(fldr.loaded))))
        self.assertEqual(len(list(fldr.not_empty)),len(fldr)-1,"not_empty attribute failed.")
        fldr-="Untitled"

    def test_methods(self):
        sliced=np.array(['DataFile', 'MDAASCIIFile', 'BNLFile', 'DataFile', 'DataFile',
       'DataFile', 'DataFile', 'MokeFile', 'EasyPlotFile', 'DataFile',
       'DataFile', 'DataFile'],
          dtype='<U12')
        fldr=SF.DataFolder(self.datadir, pattern='*.txt').sort()
        self.fldr=fldr
        test_sliced=fldr.slice_metadata("Loaded as")
        self.assertEqual(len(sliced),len(test_sliced),"Test slice not equal length - sample-data changed? {}".format(test_sliced))
        self.assertTrue(np.all(test_sliced==sliced),"Slicing metadata failed to work.")

    def test_clone(self):
         fldr=SF.DataFolder(self.datadir, pattern='*.txt')
         fldr.abc = 123 #add an attribute
         self.t = fldr.__clone__()
         self.assertTrue(self.t.pattern==fldr.pattern, 'pattern didnt copy over')
         self.assertTrue(hasattr(self.t, "abc") and self.t.abc==123, 'user attribute didnt copy over')
         self.assertTrue(isinstance(self.t['recursivefoldertest'],SF.DataFolder), 'groups didnt copy over')

    def test_grouping(self):
        fldr4=SF.DataFolder()
        x=np.linspace(-np.pi,np.pi,181)
        for phase in np.linspace(0,1.0,5):
            for amplitude in np.linspace(1,2,6):
                for frequency in np.linspace(1,2,5):
                    y=amplitude*np.sin(frequency*x+phase*np.pi)
                    d=Data(x,y,setas="xy",column_headers=["X","Y"])
                    d["frequency"]=frequency
                    d["amplitude"]=amplitude
                    d["phase"]=phase
                    d["params"]=[phase,frequency,amplitude]
                    d.filename="test/{amplitude}/{phase}/{frequency}.dat".format(**d)
                    fldr4+=d
        fldr4.unflatten()
        self.assertEqual(fldr4.mindepth,3,"Unflattened DataFolder had wrong mindepth.")
        self.assertEqual(fldr4.shape, (~~fldr4).shape,"Datafodler changed shape on flatten/unflatten")
        fldr5=fldr4.select(amplitude=1.4,recurse=True)
        fldr5.prune()
        pruned=(0,
                {'test': (0,
                   {'1.4': (0,
                     {'0.0': (5, {}),
                      '0.25': (5, {}),
                      '0.5': (5, {}),
                      '0.75': (5, {}),
                      '1.0': (5, {})})})})
        selected=(0,
                {'test': (0,
                   {'1.4': (0,
                     {'0.25': (1, {}), '0.5': (1, {}), '0.75': (1, {}), '1.0': (1, {})})})})
        self.assertEqual(fldr5.shape,pruned,"Folder pruning gave an unxpected shape.")
        self.assertEqual(fldr5[("test","1.4","0.5",0,"phase")],0.5,"Multilevel indexing of tree failed.")
        shape=(~(~fldr4).select(amplitude=1.4).select(frequency=1).select(phase__gt=0.2)).shape
        self.fldr4=fldr4
        self.assertEqual(shape, selected,"Multi selects and inverts failed.")
        g=(~fldr4)/10
        self.assertEqual(g.shape,(0,{'Group 0': (15, {}),'Group 1': (15, {}),'Group 2': (15, {}),'Group 3': (15, {}),'Group 4': (15, {}),
                                     'Group 5': (15, {}),'Group 6': (15, {}),'Group 7': (15, {}),'Group 8': (15, {}),'Group 9': (15, {})}),"Dive by int failed.")
        g["Group 6"]-=5
        self.assertEqual(g.shape,(0,{'Group 0': (15, {}),'Group 1': (15, {}),'Group 2': (15, {}),'Group 3': (15, {}),'Group 4': (15, {}),
                                     'Group 5': (15, {}),'Group 6': (14, {}),'Group 7': (15, {}),'Group 8': (15, {}),'Group 9': (15, {})}),"Sub by int failed.")
        remove=g["Group 3"][4]
        g["Group 3"]-=remove
        self.assertEqual(g.shape,(0,{'Group 0': (15, {}),'Group 1': (15, {}),'Group 2': (15, {}),'Group 3': (14, {}),'Group 4': (15, {}),
                                     'Group 5': (15, {}),'Group 6': (14, {}),'Group 7': (15, {}),'Group 8': (15, {}),'Group 9': (15, {})}),"Sub by object failed.")
        d=fldr4["test",1.0,1.0].gather(0,1)
        self.assertEqual(d.shape,(181,6),"Gather seems have failed.")
        self.assertTrue(np.all(fldr4["test",1.0,1.0].slice_metadata("phase")==
                               np.ones(5)),"Slice metadata failure.")
        d=(~fldr4).extract("phase","frequency","amplitude","params")
        self.assertEqual(d.shape,(150,6),"Extract failed to produce data of correct shape.")
        self.assertEqual(d.column_headers,['phase', 'frequency', 'amplitude', 'params', 'params', 'params'],"Exctract failed to get correct column headers.")
        p=fldr4["test",1.0,1.0]
        p=SF.PlotFolder(p)
        p.plot()
        self.assertEqual(len(plt.get_fignums()),1,"Failed to generate a single plot for PlotFolder.")
        plt.close("all")

    def test_saving(self):
        fldr4=SF.DataFolder()
        x=np.linspace(-np.pi,np.pi,181)
        for phase in np.linspace(0,1.0,5):
            for amplitude in np.linspace(1,2,6):
                for frequency in np.linspace(1,2,5):
                    y=amplitude*np.sin(frequency*x+phase*np.pi)
                    d=Data(x,y,setas="xy",column_headers=["X","Y"])
                    d["frequency"]=frequency
                    d["amplitude"]=amplitude
                    d["phase"]=phase
                    d["params"]=[phase,frequency,amplitude]
                    d.filename="test/{amplitude}/{phase}/{frequency}.dat".format(**d)
                    fldr4+=d
        fldr4.unflatten()
        newdir=tempfile.mkdtemp()
        fldr4.save(newdir)
        fldr5=SF.DataFolder(newdir)
        self.assertEqual(fldr4.shape,fldr5.shape,"Saved DataFolder and loaded DataFolder have different shapes")



if __name__=="__main__": # Run some tests manually to allow debugging
    test=Folders_test("test_Folders")
    test.setUp()
    unittest.main()
    #test.test_saving()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Base classes for the Stoner package
"""
__all__ = ["_evaluatable", "regexpDict", "typeHintedDict", "metadataObject"]
from collections import OrderedDict, MutableMapping, Mapping
import re
import copy
import numpy as np
from numpy import isnan
from numpy import NaN

from ..compat import python_v3, string_types, int_types, _pattern_type
from ..tools import isiterable, isComparable

try:
    assert not python_v3  # blist doesn't seem entirely reliable in 3.5 :-(
    from blist import sorteddict
except (AssertionError, ImportError):  # Fail if blist not present or Python 3
    sorteddict = OrderedDict


class _evaluatable(object):

    """Just a placeholder to indicate that special action needs to be taken to convert a string representation to a valid Python type."""

    pass


class regexpDict(sorteddict):

    """An ordered dictionary that permits looks up by regular expression."""

    allowed_keys = (object,)

    def __lookup__(self, name, multiple=False, exact=False):
        """Lookup name and find a matching key or raise KeyError.

        Parameters:
            name (str, _pattern_type): The name to be searched for

        Keyword Arguments:
            multiple (bool): Return a singl entry ()default, False) or multiple entries

        Returns:
            Canonical key matching the specified name.

        Raises:
            KeyError: if no key matches name.
        """
        ret = None
        try:  # name directly as key
            super(regexpDict, self).__getitem__(name)
            ret = name
        except (KeyError, TypeError):  # Fall back to regular expression lookup
            if not exact and not isinstance(name, string_types + int_types):
                name = repr(name)
            if exact:
                raise KeyError("{} not a key and exact match requested.".format(name))
            nm = name
            if isinstance(name, string_types):
                try:
                    nm = re.compile(name)
                except Exception:
                    pass
            elif isinstance(name, int_types):  # We can do this because we're an OrderedDict!
                try:
                    ret = sorted(self.keys())[name]
                except IndexError:
                    raise KeyError("{} is not a match to any key.".format(name))
            else:
                nm = name
            if isinstance(nm, _pattern_type):
                ret = [n for n in self.keys() if isinstance(n, string_types) and nm.match(n)]
        if ret is None or isiterable(ret) and not ret:
            raise KeyError("{} is not a match to any key.".format(name))
        else:
            if multiple:  # sort out returing multiple entries or not
                if not isinstance(ret, list):
                    ret = [ret]
            else:
                if isinstance(ret, list):
                    ret = ret[0]
            return ret

    def __getitem__(self, name):
        """Adds a lookup via regular expression when retrieving items."""
        return super(regexpDict, self).__getitem__(self.__lookup__(name))

    def __setitem__(self, name, value):
        """Overwrites any matching key, or if not found adds a new key."""
        try:
            key = self.__lookup__(name, exact=True)
        except KeyError:
            if not isinstance(name, self.allowed_keys):
                raise KeyError("{} is not a match to any key.".format(name))
            key = name
        super(regexpDict, self).__setitem__(key, value)

    def __delitem__(self, name):
        """Deletes keys that match by regular expression as well as exact matches"""
        super(regexpDict, self).__delitem__(self.__lookup__(name))

    def __contains__(self, name):
        """Returns True if name either is an exact key or matches when interpreted as a regular experssion."""
        try:
            name = self.__lookup__(name)
            return True
        except (KeyError, TypeError):
            return False

    def __eq__(self, other):
        """Define equals operation in terms of xor operation."""
        return len(self ^ other) == 0 and len(other ^ self) == 0

    def __xor__(self, other):
        """Give the difference between two arrays."""
        if not isinstance(other, Mapping):
            return NotImplemented
        mk = set(self.keys())
        ok = set(other.keys())
        if mk != ok:  # Keys differ
            return mk ^ ok
        # Do values differ?
        ret = OrderedDict()
        for (mk, mv), (ok, ov) in zip(sorted(self.items()), sorted(other.items())):
            if np.any(mv != ov) and isComparable(mv, ov):
                ret[mk] = (mv, ov)
        return ret

    def has_key(self, name):
        """Key is definitely in dictionary as literal"""
        return super(regexpDict, self).__contains__(name)


class typeHintedDict(regexpDict):

    """Extends a :py:class:`blist.sorteddict` to include type hints of what each key contains.

    The CM Physics Group at Leeds makes use of a standard file format that closely matches
    the :py:class:`DataFile` data structure. However, it is convenient for this file format
    to be ASCII text for ease of use with other programs. In order to represent metadata which
    can have arbitary types, the LabVIEW code that generates the data file from our measurements
    adds a type hint string. The Stoner Python code can then make use of this type hinting to
    choose the correct representation for the metadata. The type hinting information is retained
    so that files output from Python will retain type hints to permit them to be loaded into
    strongly typed languages (sch as LabVIEW).

    Attributes:
        _typehints (dict): The backing store for the type hint information
        __regexGetType (re): Used to extract the type hint from a string
        __regexSignedInt (re): matches type hint strings for signed intergers
        __regexUnsignedInt (re): matches the type hint string for unsigned integers
        __regexFloat (re): matches the type hint strings for floats
        __regexBoolean (re): matches the type hint string for a boolean
        __regexStrng (re): matches the type hint string for a string variable
        __regexEvaluatable (re): matches the type hint string for a compoind data type
        __types (dict): mapping of type hinted types to actual Python types
        __tests (dict): mapping of the regex patterns to actual python types

    Notes:
        Rather than subclassing a plain dict, this is a subclass of a :py:class:`blist.sorteddict` which stores the entries in a binary list structure.
        This makes accessing the keys much faster and also ensures that keys are always returned in alphabetical order.
    """

    allowed_keys = string_types
    # Force metadata keys to be strings
    _typehints = sorteddict()

    __regexGetType = re.compile(r"([^\{]*)\{([^\}]*)\}")
    # Match the contents of the inner most{}
    __regexSignedInt = re.compile(r"^I\d+")
    # Matches all signed integers
    __regexUnsignedInt = re.compile(r"^U / d+")
    # Match unsigned integers
    __regexFloat = re.compile(r"^(Extended|Double|Single)\sFloat")
    # Match floating point types
    __regexBoolean = re.compile(r"^Boolean")
    __regexString = re.compile(r"^(String|Path|Enum)")
    __regexEvaluatable = re.compile(r"^(Cluster||\d+D Array|List)")

    __types = OrderedDict(
        [  # Key order does matter here!
            ("Boolean", bool),
            ("I32", int),
            ("Double Float", float),
            ("Cluster", dict),
            ("AnonCluster", tuple),
            ("Array", np.ndarray),
            ("List", list),
            ("String", str),
        ]
    )
    # This is the inverse of the __tests below - this gives
    # the string type for standard Python classes

    __tests = [
        (__regexSignedInt, int),
        (__regexUnsignedInt, int),
        (__regexFloat, float),
        (__regexBoolean, bool),
        (__regexString, str),
        (__regexEvaluatable, _evaluatable()),
    ]

    # This is used to work out the correct python class for
    # some string types

    def __init__(self, *args, **kargs):
        """Construct the typeHintedDict.

        Args:
            *args, **kargs: Pass any parameters through to the dict() constructor.


        Calls the dict() constructor, then runs through the keys of the
        created dictionary and either uses the string type embedded in
        the keyname to generate the type hint (and remove the
        embedded string type from the keyname) or determines the likely
        type hint from the value of the dict element.
        """
        super(typeHintedDict, self).__init__(*args, **kargs)
        for key in list(self.keys()):  # Chekc through all the keys and see if they contain
            # type hints. If they do, move them to the
            # _typehint dict
            value = super(typeHintedDict, self).__getitem__(key)
            super(typeHintedDict, self).__delitem__(key)
            self[key] = value  # __Setitem__ has the logic to handle embedded type hints correctly

    @property
    def types(self):
        """Return the dictrionary of value types."""
        return self._typehints

    def findtype(self, value):
        """Determines the correct string type to return for common python classes.

        Args:
            value (any): The data value to determine the type hint for.

        Returns:
            A type hint string

        Note:
            Understands booleans, strings, integers, floats and np
            arrays(as arrays), and dictionaries (as clusters).
        """
        typ = "String"
        for t in self.__types:
            if isinstance(value, self.__types[t]):
                if t == "Cluster" or t == "AnonCluster":
                    elements = []
                    if isinstance(value, dict):
                        for k in value:
                            elements.append(self.findtype(value[k]))
                    else:
                        for v in value:
                            elements.append(self.findtype(v))
                    tt = ","
                    tt = tt.join(elements)
                    typ = "Cluster (" + tt + ")"
                elif t == "Array":
                    z = np.zeros(1, dtype=value.dtype)
                    typ = "{}D Array ({})".format(len(np.shape(value)), self.findtype(z[0]))
                else:
                    typ = t
                break
        return typ

    def __mungevalue(self, t, value):
        """Based on a string type t, return value cast to an appropriate python class.

        Args:
            t (string): is a string representing the type
            value (any): is the data value to be munged into the
                        correct class
        Returns:
            Returns the munged data value

        Detail:
            The class has a series of precompiled regular
            expressions that will match type strings, a list of these has been
            constructed with instances of the matching Python classes. These
            are tested in turn and if the type string matches the constructor of
            the associated python class is called with value as its argument.
        """
        ret = None
        for (regexp, valuetype) in self.__tests:
            m = regexp.search(t)
            if m is not None:
                if isinstance(valuetype, _evaluatable):
                    try:
                        array = np.array
                        nan = np.nan
                        if isinstance(value, string_types):  # we've got a string already don't need repr
                            ret = eval(value, globals(), locals())
                        else:
                            ret = eval(repr(value), globals(), locals())  # pylint: disable=eval-used
                    except NameError:
                        ret = str(value)
                    except SyntaxError:
                        ret = ""
                    break
                else:
                    ret = valuetype(value)
                    break
        else:
            ret = str(value)
        return ret

    def string_to_type(self, value):
        """Given a string value try to work out if there is a better python type dor the value.

        First of all the first character is checked to see if it is a [ or { which would
        suggest this is a list of dictionary. If the value looks like a common boolean
        value (i.e. Yes, No, True, Fale, On, Off) then it is assumed to be a boolean value.
        Fianlly it interpretation as an int, float or string is tried.

        Args:
            value (string): string representation of he value
        Returns:
            A python object of the natural type for value
        """
        ret = None
        if not isinstance(value, string_types):
            raise TypeError("Value must be a string not a {}".format(type(value)))
        value = value.strip()
        if value:
            tests = ["list(" + value + ")", "dict(" + value + ")"]
            try:
                i = "[{".index(value[0])
                array = np.array
                nan = np.nan
                ret = eval(tests[i], globals(), locals())  # pylint: disable=eval-used
            except (SyntaxError, ValueError):
                if value.lower() in ["true", "yes", "on", "false", "no", "off"]:
                    ret = value.lower() in ["true", "yes", "on"]  # Boolean
                else:
                    for trial in [int, float, str]:
                        try:
                            ret = trial(value)
                            break
                        except ValueError:
                            continue
                    else:
                        ret = None
        return ret

    def __deepcopy__(self, memo):
        """Implements a deepcopy method for typeHintedDict to work around something that gives a hang in newer Python 2.7.x"""
        cls = self.__class__
        result = cls()
        memo[id(self)] = result
        for k in self:
            result[k] = copy.deepcopy(self[k])
            result.types[k] = self.types[k]
        return result

    def _get_name_(self, name):
        """Checks a string name for an embedded type hint and strips it out.

        Args:
            name(string): String containing the name with possible type hint embedeed
        Returns:
            (name,typehint) (tuple): A tuple containing just the name of the mateadata and (if found
                the type hint string),
        """
        search = str(name)
        m = self.__regexGetType.search(search)
        if m is not None:
            return m.group(1), m.group(2)
        elif not isinstance(name, string_types + int_types):
            return search, None
        else:
            return name, None

    def __getitem__(self, name):
        """Provides a get item method that checks whether its been given a typehint in the item name and deals with it appropriately.

        Args:
            name (string): metadata key to retrieve

        Returns:
            metadata value
        """
        key = name
        (name, typehint) = self._get_name_(name)
        name = self.__lookup__(name, True)
        value = [super(typeHintedDict, self).__getitem__(nm) for nm in name]
        if typehint is not None:
            value = [self.__mungevalue(typehint, v) for v in value]
        if len(value) == 0:  # pylint: disable=len-as-condition
            raise KeyError("{} is not a valid key even when interpreted as a sregular expression!".format(key))
        elif len(value) == 1:
            return value[0]
        else:
            return {k: v for k, v in zip(name, value)}

    def __setitem__(self, name, value):
        """Provides a method to set an item in the dict, checking the key for an embedded type hint or inspecting the value as necessary.

        Args:
            name (string): The metadata keyname
            value (any): The value to store in the metadata string

        Note:
            If you provide an embedded type string it is your responsibility
            to make sure that it correctly describes the actual data
            typehintDict does not verify that your data and type string are
            compatible.
        """
        name, typehint = self._get_name_(name)
        if typehint is not None:
            self._typehints[name] = typehint
            if value is None:  # Empty data so reset to string and set empty #RCT changed the test here
                super(typeHintedDict, self).__setitem__(name, "")
                self._typehints[name] = "String"
            else:
                try:
                    super(typeHintedDict, self).__setitem__(name, self.__mungevalue(typehint, value))
                except ValueError:
                    pass  # Silently fail
        else:
            self._typehints[name] = self.findtype(value)
            super(typeHintedDict, self).__setitem__(name, value)

    def __delitem__(self, name):
        """Deletes the specified key.

        Args:
            name (string): The keyname to be deleted
        """
        name = self._get_name_(name)[0]
        name = self.__lookup__(name)

        del self._typehints[name]
        super(typeHintedDict, self).__delitem__(name)

    def __repr__(self):
        """Create a text representation of the dictionary with type data."""
        ret = ["{}:{}:{}".format(repr(key), self.type(key), repr(self[key])) for key in sorted(self)]
        return "\n".join(ret)

    def copy(self):
        """Provides a copy method that is aware of the type hinting strings.

        This produces a flat dictionary with the type hint embedded in the key name.

        Returns:
            A copy of the current typeHintedDict
        """
        cls = self.__class__
        ret = cls()
        for k in self.keys():
            t = self._typehints[k]
            ret._typehints[k] = t
            super(typeHintedDict, ret).__setitem__(k, copy.deepcopy(self[k]))
        return ret

    def filter(self, name):
        """Filter the dictionary keys by name

        Reduce the metadata dictionary leaving only keys satisfied by name.

        Keyword Arguments:
            name(str or callable):
                either a str to match or a callable function that takes metadata key-value
                as an argument and returns True or False
        """
        rem = []
        for k in self.keys():
            if isinstance(name, string_types):
                if not name in k:
                    rem.append(k)
            elif hasattr(name, "__call__"):
                if not name(k):
                    rem.append(k)
            else:
                raise ValueError("name must be a string or a function")
        for k in rem:
            del self[k]

    def type(self, key):
        """Returns the typehint for the given k(s).

        This simply looks up the type hinting dictionary for each key it is given.

        Args:
            key (string or sequence of strings): Either a single string key or a iterable type containing
                keys
        Returns:
            The string type hint (or a list of string type hints)
        """
        if isinstance(key, string_types):
            return self._typehints[key]
        else:
            try:
                return [self._typehints[x] for x in key]
            except TypeError:
                return self._typehints[key]

    def export(self, key):
        """Exports a single metadata value to a string representation with type hint.

        In the ASCII based file format, the type hinted metadata is represented
        in the first column of a tab delimited text file as a series of lines
        with format keyname{typhint}=string_value.

        Args:
            key (string): The metadata key to export
        Returns:
            A string of the format : key{type hint} = value
        """
        if isinstance(self[key], string_types):  # avoid string within string problems and backslash overdrive
            ret = "{}{{{}}}={}".format(key, self.type(key), self[key])
        else:
            ret = "{}{{{}}}={}".format(key, self.type(key), repr(self[key]))
        return ret

    def export_all(self):
        """Return all the entries in the typeHintedDict as a list of exported lines.

        Returns:
            (list of str): A list of exported strings

        Notes:
            The keys are returned in sorted order as a result of the underlying blist.sorteddict meothd.
        """
        return [self.export(x) for x in self]

    def import_all(self, lines):
        """Reads multiple lines of strings and tries to import keys from them.

        Args:
            lines(list of str): The lines of metadata values to import.
        """
        for line in lines:
            self.import_key(line)

    def import_key(self, line):
        """Import a single key from a string like key{type hint} = value.

        This is the inverse of the :py:meth:`typeHintedDict.export` method.

        Args:
            line(str): The string line to be interpreted as a key-value pair.

        """
        parts = line.split("=")
        k = parts[0]
        v = "=".join(parts[1:])  # rejoin any = in the value string
        self[k] = v


class metadataObject(MutableMapping):

    """Provides a base class representing some sort of object that has metadata stored in a :py:class:`Stoner.Core.typeHintedDict` object.

    Attributes:
        metadata (typeHintedDict): of key-value metadata pairs. The dictionary tries to retain information about the type of data so as to aid import and
            export from CM group LabVIEW code.
   """

    def __init__(self, *args, **kargs):
        """Initialises the current metadata attribute."""
        metadata = kargs.pop("metadata", None)
        if metadata is not None:
            self.metadata.update(metadata)
        super(metadataObject, self).__init__()

    @property
    def metadata(self):
        """Read the metadata dictionary."""
        try:
            return self._metadata
        except AttributeError:  # Oops no metadata yet
            self._metadata = typeHintedDict()
            return self._metadata

    @metadata.setter
    def metadata(self, value):
        """Update the metadata object with type checking."""
        if not isinstance(value, typeHintedDict) and isiterable(value):
            self._metadata = typeHintedDict(value)
        elif isinstance(value, typeHintedDict):
            self._metadata = value
        else:
            raise TypeError(
                "metadata must be something that can be turned into a dictionary, not a {}".format(type(value))
            )

    def __getitem__(self, name):
        """Pass through to metadata dictionary."""
        return self.metadata[name]

    def __setitem__(self, name, value):
        """Pass through to metadata dictionary."""
        self.metadata[name] = value

    def __delitem__(self, name):
        """Pass through to metadata dictionary."""
        del self.metadata[name]

    def __eq__(self, other):
        """Implement am equality test for metadataObjects."""
        if not isinstance(other, metadataObject):
            return False
        if len(self) != len(other):
            return False
        ret = self.metadata ^ other.metadata
        return len(ret) == 0

    def __len__(self):
        """Pass through to metadata dictionary."""
        return len(self.metadata)

    def __iter__(self):
        """Pass through to metadata dictionary."""
        return self.metadata.__iter__()

    def keys(self):
        """Return the keys of the metadata dictionary."""
        for k in self.metadata.keys():
            yield k

    def items(self):
        """Make sure we implement an items that doesn't just iterate over self!"""
        for k, v in self.metadata.items():
            yield k, v

    def values(self):
        """Return the values of the metadata dictionary."""
        for v in self.metadata.values():
            yield v

    def save(self, path):
        """Stub method for a save function."""
        raise NotImplementedError("Save is not implemented in the base class.")

    def _load(self, filename):
        """Stub method for a load function."""
        raise NotImplementedError("Save is not implemented in the base class.")
